from slither.core.variables.state_variable import StateVariable
from slither.vyper_parsing.ast.types import VariableDecl
from slither.vyper_parsing.type_parsing import parse_type
from slither.vyper_parsing.expressions.expression_parsing import parse_expression


class StateVariableVyper:
    def __init__(self, variable: StateVariable, variable_data: VariableDecl) -> None:
        self._variable: StateVariable = variable
        self._variable.name = variable_data.target.id
        self._variable.is_constant = variable_data.is_constant
        self._variable.is_immutable = variable_data.is_immutable
        self._variable.visibility = "public" if variable_data.is_public else "internal"
        self._elem_to_parse = variable_data.annotation

        if variable_data.value is not None:
            self._variable.initialized = True
            self._initializedNotParsed = variable_data.value

    @property
    def underlying_variable(self) -> StateVariable:
        return self._variable

    def analyze(self, contract) -> None:
        self._variable.type = parse_type(self._elem_to_parse, contract)

        if self._variable.initialized:
            self._variable.expression = parse_expression(self._initializedNotParsed, contract)
            self._initializedNotParsed = None



"""
Module printing summary of the contract
"""

from slither.printers.abstract_printer import AbstractPrinter


class PrinterSlithIRSSA(AbstractPrinter):
    ARGUMENT = "slithir-ssa"
    HELP = "Print the slithIR representation of the functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#slithir-ssa"

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        for contract in self.contracts:
            txt += f"Contract {contract.name}" + "\n"
            for function in contract.functions:
                txt += f"\tFunction {function.canonical_name}" + "\n"
                for node in function.nodes:
                    if node.expression:
                        txt += f"\t\tExpression: {node.expression}" + "\n"
                    if node.irs_ssa:
                        txt += "\t\tIRs:" + "\n"
                        for ir in node.irs_ssa:
                            txt += f"\t\t\t{ir}" + "\n"
            for modifier in contract.modifiers:
                txt += f"\tModifier {modifier.canonical_name}" + "\n"
                for node in modifier.nodes:
                    txt += str(node) + "\n"
                    if node.expression:
                        txt += f"\t\tExpression: {node.expression}" + "\n"
                    if node.irs_ssa:
                        txt += "\t\tIRs:" + "\n"
                        for ir in node.irs_ssa:
                            txt += f"\t\t\t{ir}" + "\n"
        self.info(txt)
        res = self.generate_output(txt)
        return res



import logging
import re
from argparse import Namespace
from collections.abc import Callable

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import FunctionContract
from slither.core.variables.state_variable import StateVariable
from slither.utils import output
from slither.utils.colors import yellow, green, red

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("Slither.kspec")


def _refactor_type(targeted_type: str) -> str:
    return {"uint": "uint256", "int": "int256"}.get(targeted_type, targeted_type)


def _get_all_covered_kspec_functions(target: str) -> set[tuple[str, str]]:
    # Create a set of our discovered functions which are covered
    covered_functions: set[tuple[str, str]] = set()

    BEHAVIOUR_PATTERN = re.compile(r"behaviour\s+(\S+)\s+of\s+(\S+)")
    INTERFACE_PATTERN = re.compile(r"interface\s+([^\r\n]+)")

    # Read the file contents
    with open(target, encoding="utf8") as target_file:
        lines = target_file.readlines()

    # Loop for each line, if a line matches our behaviour regex, and the next one matches our interface regex,
    # we add our finding
    i = 0
    while i < len(lines):
        match = BEHAVIOUR_PATTERN.match(lines[i])
        if match:
            contract_name = match.groups()[1]
            match = INTERFACE_PATTERN.match(lines[i + 1])
            if match:
                function_full_name = match.groups()[0]
                start, end = (
                    function_full_name.index("(") + 1,
                    function_full_name.index(")"),
                )
                function_arguments = function_full_name[start:end].split(",")
                function_arguments = [
                    _refactor_type(arg.strip().split(" ")[0]) for arg in function_arguments
                ]
                function_full_name = function_full_name[:start] + ",".join(function_arguments) + ")"
                covered_functions.add((contract_name, function_full_name))
                i += 1
        i += 1
    return covered_functions


def _get_slither_functions(
    slither: SlitherCompilationUnit,
) -> dict[tuple[str, str], FunctionContract | StateVariable]:
    # Use contract == contract_declarer to avoid duplicate
    all_functions_declared: list[FunctionContract | StateVariable] = [
        f
        for f in slither.functions
        if (
            (isinstance(f, FunctionContract) and f.contract == f.contract_declarer)
            and f.is_implemented
            and not f.is_constructor
            and not f.is_constructor_variables
        )
    ]
    # Use list(set()) because same state variable instances can be shared across contracts
    # TODO: integrate state variables
    all_functions_declared += list(
        {s for s in slither.state_variables if s.visibility in ["public", "external"]}
    )
    slither_functions = {
        (function.contract.name, function.full_name): function
        for function in all_functions_declared
    }

    return slither_functions


def _generate_output(
    kspec: list[FunctionContract | StateVariable],
    message: str,
    color: Callable[[str], str],
    generate_json: bool,
) -> dict | None:
    info = ""
    for function in kspec:
        info += f"{message} {function.contract.name}.{function.full_name}\n"
    if info:
        logger.info(color(info))

    if generate_json:
        json_kspec_present = output.Output(info)
        for function in kspec:
            json_kspec_present.add(function)
        return json_kspec_present.data
    return None


def _generate_output_unresolved(
    kspec: set[tuple[str, str]], message: str, color: Callable[[str], str], generate_json: bool
) -> dict | None:
    info = ""
    for contract, function in kspec:
        info += f"{message} {contract}.{function}\n"
    if info:
        logger.info(color(info))

    if generate_json:
        json_kspec_present = output.Output(info, additional_fields={"signatures": kspec})
        return json_kspec_present.data
    return None


def _run_coverage_analysis(
    args: Namespace, slither: SlitherCompilationUnit, kspec_functions: set[tuple[str, str]]
) -> None:
    # Collect all slither functions
    slither_functions = _get_slither_functions(slither)

    # Determine which klab specs were not resolved.
    slither_functions_set = set(slither_functions)
    kspec_functions_resolved = kspec_functions & slither_functions_set
    kspec_functions_unresolved: set[tuple[str, str]] = kspec_functions - kspec_functions_resolved

    kspec_missing: list[FunctionContract | StateVariable] = []
    kspec_present: list[FunctionContract | StateVariable] = []

    for slither_func_desc in sorted(slither_functions_set):
        slither_func = slither_functions[slither_func_desc]

        if slither_func_desc in kspec_functions:
            kspec_present.append(slither_func)
        else:
            kspec_missing.append(slither_func)

    logger.info("## Check for functions coverage")
    json_kspec_present = _generate_output(kspec_present, "[✓]", green, args.json)
    json_kspec_missing_functions = _generate_output(
        [f for f in kspec_missing if isinstance(f, FunctionContract)],
        "[ ] (Missing function)",
        red,
        args.json,
    )
    json_kspec_missing_variables = _generate_output(
        [f for f in kspec_missing if isinstance(f, StateVariable)],
        "[ ] (Missing variable)",
        yellow,
        args.json,
    )
    json_kspec_unresolved = _generate_output_unresolved(
        kspec_functions_unresolved, "[ ] (Unresolved)", yellow, args.json
    )

    # Handle unresolved kspecs
    if args.json:
        output.output_to_json(
            args.json,
            None,
            {
                "functions_present": json_kspec_present,
                "functions_missing": json_kspec_missing_functions,
                "variables_missing": json_kspec_missing_variables,
                "functions_unresolved": json_kspec_unresolved,
            },
        )


def run_analysis(args: Namespace, slither: SlitherCompilationUnit, kspec_arg: str) -> None:
    # Get all of our kspec'd functions (tuple(contract_name, function_name)).
    if "," in kspec_arg:
        kspecs = kspec_arg.split(",")
        kspec_functions: set[tuple[str, str]] = set()
        for kspec in kspecs:
            kspec_functions |= _get_all_covered_kspec_functions(kspec)
    else:
        kspec_functions = _get_all_covered_kspec_functions(kspec_arg)

    # Run coverage analysis
    _run_coverage_analysis(args, slither, kspec_functions)



from slither.core.variables.local_variable import LocalVariable
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.variable import SlithIRVariable
from slither.slithir.variables.state_variable import StateIRVariable


class LocalIRVariable(LocalVariable, SlithIRVariable):
    def __init__(self, local_variable: LocalVariable) -> None:
        assert isinstance(local_variable, LocalVariable)

        super().__init__()

        # initiate ChildContract
        self.set_function(local_variable.function)

        # initiate Variable
        self._name = local_variable.name
        self._initial_expression = local_variable.expression
        self._type = local_variable.type
        self._initialized = local_variable.initialized
        self._visibility = local_variable.visibility
        self._is_constant = local_variable.is_constant

        # initiate LocalVariable
        self._location = local_variable.location
        self._is_storage = local_variable.is_storage

        self._index = 0

        # Additional field
        # points to state variables
        self._refers_to: set[StateIRVariable] = set()

        # keep un-ssa version
        if isinstance(local_variable, LocalIRVariable):
            self._non_ssa_version = local_variable.non_ssa_version
        else:
            self._non_ssa_version = local_variable

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, idx: int) -> None:
        self._index = idx

    @property
    def refers_to(self):
        """
        Return the alias for local variable that are storage pointers
        """
        if self.is_storage:
            return self._refers_to
        return set()

    @refers_to.setter
    def refers_to(self, variables):
        self._refers_to = variables

    @property
    def non_ssa_version(self) -> LocalVariable:
        return self._non_ssa_version

    def add_refers_to(self, variable: StateIRVariable) -> None:
        # It is a temporaryVariable if its the return of a new ..
        # ex: string[] memory dynargs = new string[](1);
        assert isinstance(variable, (SlithIRVariable, TemporaryVariable))
        self._refers_to.add(variable)

    @property
    def ssa_name(self):
        if self.is_storage:
            return f"{self._name}_{self.index} (-> {[v.name for v in self.refers_to]})"
        return f"{self._name}_{self.index}"



"""
Module printing the call graph

The call graph shows for each function,
what are the contracts/functions called.
The output is a dot file named filename.dot
"""

from collections import defaultdict
from collections.abc import Sequence

from slither.core.declarations import Contract, FunctionContract
from slither.core.declarations.function import Function
from slither.slithir.operations import HighLevelCall, InternalCall
from slither.core.declarations.solidity_variables import SolidityFunction
from slither.core.variables.variable import Variable
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.output import Output


def _contract_subgraph(contract: Contract) -> str:
    return f"cluster_{contract.id}_{contract.name}"


# return unique id for contract function to use as node name
def _function_node(contract: Contract, function: Function | Variable) -> str:
    return f"{contract.id}_{function.name}"


# return unique id for solidity function to use as node name
def _solidity_function_node(solidity_function: SolidityFunction) -> str:
    return f"{solidity_function.name}"


# return dot language string to add graph edge
def _edge(from_node: str, to_node: str) -> str:
    return f'"{from_node}" -> "{to_node}"'


# return dot language string to add graph node (with optional label)
def _node(node: str, label: str | None = None) -> str:
    return " ".join(
        (
            f'"{node}"',
            f'[label="{label}"]' if label is not None else "",
        )
    )


def _process_internal_call(
    contract: Contract,
    function: Function,
    internal_call: InternalCall,
    contract_calls: dict[Contract, set[str]],
    solidity_functions: set[str],
    solidity_calls: set[str],
) -> None:
    if isinstance(internal_call.function, (Function)):
        contract_calls[contract].add(
            _edge(
                _function_node(contract, function),
                _function_node(contract, internal_call.function),
            )
        )
    elif isinstance(internal_call.function, (SolidityFunction)):
        solidity_functions.add(
            _node(_solidity_function_node(internal_call.function)),
        )
        solidity_calls.add(
            _edge(
                _function_node(contract, function),
                _solidity_function_node(internal_call.function),
            )
        )


def _render_external_calls(external_calls: set[str]) -> str:
    return "\n".join(external_calls)


def _render_internal_calls(
    contract: Contract,
    contract_functions: dict[Contract, set[str]],
    contract_calls: dict[Contract, set[str]],
) -> str:
    lines = []

    lines.append(f"subgraph {_contract_subgraph(contract)} {{")
    lines.append(f'label = "{contract.name}"')

    lines.extend(contract_functions[contract])
    lines.extend(contract_calls[contract])

    lines.append("}")

    return "\n".join(lines)


def _render_solidity_calls(solidity_functions: set[str], solidity_calls: set[str]) -> str:
    lines = []

    lines.append("subgraph cluster_solidity {")
    lines.append('label = "[Solidity]"')

    lines.extend(solidity_functions)
    lines.extend(solidity_calls)

    lines.append("}")

    return "\n".join(lines)


def _process_external_call(
    contract: Contract,
    function: Function,
    external_call: tuple[Contract, HighLevelCall],
    contract_functions: dict[Contract, set[str]],
    external_calls: set[str],
    all_contracts: set[Contract],
) -> None:
    external_contract, ir = external_call

    if external_contract not in all_contracts:
        return

    # add variable as node to respective contract
    if isinstance(ir.function, (Variable)):
        contract_functions[external_contract].add(
            _node(
                _function_node(external_contract, ir.function),
                ir.function.name,
            )
        )

    external_calls.add(
        _edge(
            _function_node(contract, function),
            _function_node(external_contract, ir.function),
        )
    )


def _process_function(
    contract: Contract,
    function: Function,
    contract_functions: dict[Contract, set[str]],
    contract_calls: dict[Contract, set[str]],
    solidity_functions: set[str],
    solidity_calls: set[str],
    external_calls: set[str],
    all_contracts: set[Contract],
) -> None:
    contract_functions[contract].add(
        _node(_function_node(contract, function), function.name),
    )

    for internal_call in function.internal_calls:
        _process_internal_call(
            contract,
            function,
            internal_call,
            contract_calls,
            solidity_functions,
            solidity_calls,
        )
    for external_call in function.high_level_calls:
        _process_external_call(
            contract,
            function,
            external_call,
            contract_functions,
            external_calls,
            all_contracts,
        )


def _process_functions(functions: Sequence[Function]) -> str:
    # TODO  add support for top level function

    contract_functions: dict[Contract, set[str]] = defaultdict(
        set
    )  # contract -> contract functions nodes
    contract_calls: dict[Contract, set[str]] = defaultdict(set)  # contract -> contract calls edges

    solidity_functions: set[str] = set()  # solidity function nodes
    solidity_calls: set[str] = set()  # solidity calls edges
    external_calls: set[str] = set()  # external calls edges

    all_contracts = set()

    for function in functions:
        if isinstance(function, FunctionContract):
            all_contracts.add(function.contract_declarer)
    for function in functions:
        if isinstance(function, FunctionContract):
            _process_function(
                function.contract_declarer,
                function,
                contract_functions,
                contract_calls,
                solidity_functions,
                solidity_calls,
                external_calls,
                all_contracts,
            )

    render_internal_calls = ""
    for contract in all_contracts:
        render_internal_calls += _render_internal_calls(
            contract, contract_functions, contract_calls
        )

    render_solidity_calls = _render_solidity_calls(solidity_functions, solidity_calls)

    render_external_calls = _render_external_calls(external_calls)

    return render_internal_calls + render_solidity_calls + render_external_calls


class PrinterCallGraph(AbstractPrinter):
    ARGUMENT = "call-graph"
    HELP = "Export the call-graph of the contracts to a dot file"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#call-graph"

    def output(self, filename: str) -> Output:
        """
        Output the graph in filename
        Args:
            filename(string)
        """

        all_contracts_filename = ""
        if not filename.endswith(".dot"):
            if filename in ("", "."):
                filename = ""
            else:
                filename += "."
            all_contracts_filename = f"{filename}all_contracts.call-graph.dot"

        if filename == ".dot":
            all_contracts_filename = "all_contracts.dot"

        info = ""
        results = []
        with open(all_contracts_filename, "w", encoding="utf8") as f:
            info += f"Call Graph: {all_contracts_filename}\n"

            # Avoid duplicate functions due to different compilation unit
            all_functionss = [
                compilation_unit.functions for compilation_unit in self.slither.compilation_units
            ]
            all_functions = [item for sublist in all_functionss for item in sublist]
            all_functions_as_dict = {
                function.canonical_name: function for function in all_functions
            }
            content = "\n".join(
                ["strict digraph {"]
                + ['rankdir="LR"']
                + ["node [shape=box]"]
                + [_process_functions(list(all_functions_as_dict.values()))]
                + ["}"]
            )
            f.write(content)
            results.append((all_contracts_filename, content))

        for derived_contract in self.slither.contracts_derived:
            derived_output_filename = f"{filename}{derived_contract.name}.call-graph.dot"
            with open(derived_output_filename, "w", encoding="utf8") as f:
                info += f"Call Graph: {derived_output_filename}\n"
                content = "\n".join(
                    ["strict digraph {"]
                    + ['rankdir="LR"']
                    + ["node [shape=box]"]
                    + [_process_functions(derived_contract.functions)]
                    + ["}"]
                )
                f.write(content)
                results.append((derived_output_filename, content))

        self.info(info)
        res = self.generate_output(info)
        for filename_result, content in results:
            res.add_file(filename_result, content)

        return res



from slither.slithir.operations.lvalue import OperationWithLValue
from slither.core.solidity_types.elementary_type import ElementaryType


class TmpNewElementaryType(OperationWithLValue):
    def __init__(self, new_type: ElementaryType, lvalue):
        assert isinstance(new_type, ElementaryType)
        super().__init__()
        self._type: ElementaryType = new_type
        self._lvalue = lvalue

    @property
    def read(self) -> list:
        return []

    @property
    def type(self) -> ElementaryType:
        return self._type

    def __str__(self) -> str:
        return f"{self.lvalue} = new {self._type}"



from enum import Enum


from typing import Any


class ComparableEnum(Enum):
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ComparableEnum):
            return self.value == other.value
        return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, ComparableEnum):
            return self.value != other.value
        return False

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, ComparableEnum):
            return self.value < other.value
        return False

    def __repr__(self) -> str:
        return f"{self.value!s}"

    def __hash__(self) -> int:
        return hash(self.value)



from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_rvalue, RVALUE
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


class InitArray(OperationWithLValue):
    def __init__(
        self, init_values: list[RVALUE], lvalue: TemporaryVariableSSA | TemporaryVariable
    ) -> None:
        # init_values can be an array of n dimension
        # reduce was removed in py3
        super().__init__()

        def reduce(xs):
            result = True
            for i in xs:
                result = result and i
            return result

        def check(elem):
            if isinstance(elem, (list,)):
                return reduce(elem)
            return is_valid_rvalue(elem)

        assert check(init_values)
        self._init_values = init_values
        self._lvalue = lvalue

    @property
    def read(self) -> list[RVALUE]:
        return self._unroll(self.init_values)

    @property
    def init_values(self) -> list[RVALUE]:
        return list(self._init_values)

    def __str__(self):
        def convert(elem):
            if isinstance(elem, (list,)):
                return str([convert(x) for x in elem])
            return f"{elem}({elem.type})"

        init_values = convert(self.init_values)
        return f"{self.lvalue}({self.lvalue.type}) = {init_values}"



from typing import TYPE_CHECKING

from slither.core.source_mapping.source_mapping import SourceMapping

if TYPE_CHECKING:
    from slither.core.declarations import Contract


class ContractLevel(SourceMapping):
    """
    This class is used to represent objects that are at the contract level
    The opposite is TopLevel

    """

    def __init__(self) -> None:
        super().__init__()
        # TODO remove all the setters for the child objects
        # And make it a constructor argument
        # This will remove the optional
        self._contract: Contract | None = None

    def set_contract(self, contract: "Contract") -> None:
        self._contract = contract

    @property
    def contract(self) -> "Contract":
        assert self._contract
        return self._contract



from pathlib import Path

from slither import Slither
from slither.printers.guidance.echidna import _extract_constants, ConstantValue

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_enum_max_min(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.19")
    slither = Slither(Path(TEST_DATA_DIR, "constantfolding.sol").as_posix(), solc=solc_path)

    contracts = slither.get_contract_from_name("A")

    constants = _extract_constants(contracts)[0]["A"]["use()"]

    assert set(constants) == {
        ConstantValue(value="2", type="uint256"),
        ConstantValue(value="10", type="uint256"),
        ConstantValue(value="100", type="uint256"),
        ConstantValue(value="4294967295", type="uint32"),
    }



from collections import defaultdict
import json
import logging
import os
import re
from pathlib import Path

from slither.analyses.data_dependency.data_dependency import compute_dependency
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import Contract, Function
from slither.core.declarations.custom_error_top_level import CustomErrorTopLevel
from slither.core.declarations.enum_top_level import EnumTopLevel
from slither.core.declarations.event_top_level import EventTopLevel
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.import_directive import Import
from slither.core.declarations.pragma_directive import Pragma
from slither.core.declarations.structure_top_level import StructureTopLevel
from slither.core.declarations.using_for_top_level import UsingForTopLevel
from slither.core.scope.scope import FileScope
from slither.core.solidity_types import ElementaryType, TypeAliasTopLevel
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.exceptions import SlitherException
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.declarations.contract import ContractSolc
from slither.solc_parsing.declarations.custom_error import CustomErrorSolc
from slither.solc_parsing.declarations.function import FunctionSolc
from slither.solc_parsing.declarations.event_top_level import EventTopLevelSolc
from slither.solc_parsing.declarations.structure_top_level import StructureTopLevelSolc
from slither.solc_parsing.declarations.using_for_top_level import UsingForTopLevelSolc
from slither.solc_parsing.exceptions import VariableNotFound
from slither.solc_parsing.variables.top_level_variable import TopLevelVariableSolc

logging.basicConfig()
logger = logging.getLogger("SlitherSolcParsing")
logger.setLevel(logging.INFO)


class InheritanceResolutionError(SlitherException):
    pass


def _handle_import_aliases(
    symbol_aliases: dict, import_directive: Import, scope: FileScope
) -> None:
    """
    Handle the parsing of import aliases

    Args:
        symbol_aliases (Dict): json dict from solc
        import_directive (Import): current import directive
        scope (FileScope): current file scape

    Returns:

    """
    for symbol_alias in symbol_aliases:
        if "foreign" in symbol_alias and "local" in symbol_alias:
            if isinstance(symbol_alias["foreign"], dict) and "name" in symbol_alias["foreign"]:
                original_name = symbol_alias["foreign"]["name"]
                local_name = symbol_alias["local"]
                import_directive.renaming[local_name] = original_name
                # Assuming that two imports cannot collide in renaming
                scope.renaming[local_name] = original_name

            # This path should only be hit for the malformed AST of solc 0.5.12 where
            # the foreign identifier cannot be found but is required to resolve the alias.
            # see https://github.com/crytic/slither/issues/1319
            elif symbol_alias["local"]:
                raise SlitherException(
                    "Cannot resolve local alias for import directive due to malformed AST. Please upgrade to solc 0.6.0 or higher."
                )


class SlitherCompilationUnitSolc(CallerContextExpression):
    def __init__(self, compilation_unit: SlitherCompilationUnit) -> None:
        super().__init__()

        self._compilation_unit: SlitherCompilationUnit = compilation_unit

        self._contracts_by_id: dict[int, Contract] = {}
        # For top level functions, there should only be one `Function` since they can't be virtual and therefore can't be overridden.
        self._functions_by_id: dict[int, list[Function]] = defaultdict(list)
        self.imports_by_id: dict[int, Import] = {}
        self.top_level_events_by_id: dict[int, EventTopLevel] = {}
        self.top_level_errors_by_id: dict[int, EventTopLevel] = {}
        self.top_level_structures_by_id: dict[int, StructureTopLevel] = {}
        self.top_level_variables_by_id: dict[int, TopLevelVariable] = {}
        self.top_level_type_aliases_by_id: dict[int, TypeAliasTopLevel] = {}
        self.top_level_enums_by_id: dict[int, EnumTopLevel] = {}

        self._parsed = False
        self._analyzed = False
        self._is_compact_ast = False

        self._underlying_contract_to_parser: dict[Contract, ContractSolc] = {}
        self._structures_top_level_parser: list[StructureTopLevelSolc] = []
        self._custom_error_parser: list[CustomErrorSolc] = []
        self._variables_top_level_parser: list[TopLevelVariableSolc] = []
        self._functions_top_level_parser: list[FunctionSolc] = []
        self._using_for_top_level_parser: list[UsingForTopLevelSolc] = []
        self._events_top_level_parser: list[EventTopLevelSolc] = []
        self._all_functions_and_modifier_parser: list[FunctionSolc] = []

        self._top_level_contracts_counter = 0

    @property
    def compilation_unit(self) -> SlitherCompilationUnit:
        return self._compilation_unit

    @property
    def all_functions_and_modifiers_parser(self) -> list[FunctionSolc]:
        return self._all_functions_and_modifier_parser

    def add_function_or_modifier_parser(self, f: FunctionSolc) -> None:
        self._all_functions_and_modifier_parser.append(f)
        self._functions_by_id[f.underlying_function.id].append(f.underlying_function)

    @property
    def underlying_contract_to_parser(self) -> dict[Contract, ContractSolc]:
        return self._underlying_contract_to_parser

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self

    @property
    def contracts_by_id(self) -> dict[int, Contract]:
        return self._contracts_by_id

    @property
    def functions_by_id(self) -> dict[int, list[Function]]:
        return self._functions_by_id

    ###################################################################################
    ###################################################################################
    # region AST
    ###################################################################################
    ###################################################################################

    def get_key(self) -> str:
        if self._is_compact_ast:
            return "nodeType"
        return "name"

    def get_children(self) -> str:
        if self._is_compact_ast:
            return "nodes"
        return "children"

    @property
    def is_compact_ast(self) -> bool:
        return self._is_compact_ast

    # endregion
    ###################################################################################
    ###################################################################################
    # region Parsing
    ###################################################################################
    ###################################################################################

    def parse_top_level_from_json(self, json_data: str) -> bool:
        try:
            data_loaded = json.loads(json_data)
            # Truffle AST
            if "ast" in data_loaded:
                self.parse_top_level_items(data_loaded["ast"], data_loaded["sourcePath"])
                return True
            # solc AST, where the non-json text was removed
            if "attributes" in data_loaded:
                filename = data_loaded["attributes"]["absolutePath"]
            else:
                filename = data_loaded["absolutePath"]
            self.parse_top_level_items(data_loaded, filename)
            return True
        except ValueError:
            first = json_data.find("{")
            if first != -1:
                last = json_data.rfind("}") + 1
                filename = json_data[0:first]
                json_data = json_data[first:last]

                data_loaded = json.loads(json_data)
                self.parse_top_level_items(data_loaded, filename)
                return True
            return False

    def _parse_enum(self, top_level_data: dict, filename: str) -> None:
        if self.is_compact_ast:
            name = top_level_data["name"]
            canonicalName = top_level_data["canonicalName"]
        else:
            name = top_level_data["attributes"][self.get_key()]
            if "canonicalName" in top_level_data["attributes"]:
                canonicalName = top_level_data["attributes"]["canonicalName"]
            else:
                canonicalName = name
        values = []
        children = (
            top_level_data["members"]
            if "members" in top_level_data
            else top_level_data.get("children", [])
        )
        for child in children:
            assert child[self.get_key()] == "EnumValue"
            if self.is_compact_ast:
                values.append(child["name"])
            else:
                values.append(child["attributes"][self.get_key()])

        scope = self.compilation_unit.get_scope(filename)
        enum = EnumTopLevel(name, canonicalName, values, scope)
        enum.set_offset(top_level_data["src"], self._compilation_unit)
        self._compilation_unit.enums_top_level.append(enum)
        scope.enums[name] = enum
        refId = top_level_data["id"]
        self.top_level_enums_by_id[refId] = enum

    def parse_top_level_items(self, data_loaded: dict, filename: str) -> None:
        if not data_loaded or data_loaded is None:
            logger.error(
                "crytic-compile returned an empty AST. "
                "If you are trying to analyze a contract from etherscan or similar make sure it has source code available."
            )
            return

        exported_symbols = {}
        if "nodeType" in data_loaded:
            self._is_compact_ast = True
            exported_symbols = data_loaded.get("exportedSymbols", {})
        else:
            attributes = data_loaded.get("attributes", {})
            exported_symbols = attributes.get("exportedSymbols", {})

        if "sourcePaths" in data_loaded:
            for sourcePath in data_loaded["sourcePaths"]:
                if os.path.isfile(sourcePath):
                    self._compilation_unit.core.add_source_code(sourcePath)

        if data_loaded[self.get_key()] == "root":
            logger.error("solc <0.4 is not supported")
            return
        if data_loaded[self.get_key()] == "SourceUnit":
            self._parse_source_unit(data_loaded, filename)
        else:
            logger.error("solc version is not supported")
            return

        if self.get_children() not in data_loaded:
            return

        scope = self.compilation_unit.get_scope(filename)
        # Exported symbols includes a reference ID to all top-level definitions the file exports,
        # including def's brought in by imports (even transitively) and def's local to the file.
        for refId in exported_symbols.values():
            scope.exported_symbols |= set(refId)

        for top_level_data in data_loaded[self.get_children()]:
            if top_level_data[self.get_key()] == "ContractDefinition":
                contract = Contract(self._compilation_unit, scope)
                contract_parser = ContractSolc(self, contract, top_level_data)
                scope.contracts[contract.name] = contract
                if "src" in top_level_data:
                    contract.set_offset(top_level_data["src"], self._compilation_unit)

                self._underlying_contract_to_parser[contract] = contract_parser

            elif top_level_data[self.get_key()] == "PragmaDirective":
                if self._is_compact_ast:
                    pragma = Pragma(top_level_data["literals"], scope)
                    scope.pragmas.add(pragma)
                else:
                    pragma = Pragma(top_level_data["attributes"]["literals"], scope)
                    scope.pragmas.add(pragma)
                pragma.set_offset(top_level_data["src"], self._compilation_unit)
                self._compilation_unit.pragma_directives.append(pragma)

            elif top_level_data[self.get_key()] == "UsingForDirective":
                scope = self.compilation_unit.get_scope(filename)
                usingFor = UsingForTopLevel(scope)
                usingFor_parser = UsingForTopLevelSolc(usingFor, top_level_data, self)
                usingFor.set_offset(top_level_data["src"], self._compilation_unit)
                scope.using_for_directives.add(usingFor)

                self._compilation_unit.using_for_top_level.append(usingFor)
                self._using_for_top_level_parser.append(usingFor_parser)

            elif top_level_data[self.get_key()] == "ImportDirective":
                referenceId = top_level_data["id"]
                if self.is_compact_ast:
                    import_directive = Import(
                        Path(
                            top_level_data["absolutePath"],
                        ),
                        scope,
                    )
                    scope.imports.add(import_directive)
                    # TODO investigate unitAlias in version < 0.7 and legacy ast
                    if "unitAlias" in top_level_data:
                        import_directive.alias = top_level_data["unitAlias"]
                    if "symbolAliases" in top_level_data:
                        symbol_aliases = top_level_data["symbolAliases"]
                        _handle_import_aliases(symbol_aliases, import_directive, scope)
                else:
                    import_directive = Import(
                        Path(
                            top_level_data["attributes"].get("absolutePath", ""),
                        ),
                        scope,
                    )
                    scope.imports.add(import_directive)
                    # TODO investigate unitAlias in version < 0.7 and legacy ast
                    if (
                        "attributes" in top_level_data
                        and "unitAlias" in top_level_data["attributes"]
                    ):
                        import_directive.alias = top_level_data["attributes"]["unitAlias"]
                import_directive.set_offset(top_level_data["src"], self._compilation_unit)
                self._compilation_unit.import_directives.append(import_directive)
                self.imports_by_id[referenceId] = import_directive

                get_imported_scope = self.compilation_unit.get_scope(import_directive.filename)
                scope.accessible_scopes.append(get_imported_scope)

            elif top_level_data[self.get_key()] == "StructDefinition":
                st = StructureTopLevel(self.compilation_unit, scope)
                st.set_offset(top_level_data["src"], self._compilation_unit)
                st_parser = StructureTopLevelSolc(st, top_level_data, self)
                scope.structures[st.name] = st

                self._compilation_unit.structures_top_level.append(st)
                self._structures_top_level_parser.append(st_parser)
                referenceId = top_level_data["id"]
                self.top_level_structures_by_id[referenceId] = st

            elif top_level_data[self.get_key()] == "EnumDefinition":
                # Note enum don't need a complex parser, so everything is directly done
                self._parse_enum(top_level_data, filename)

            elif top_level_data[self.get_key()] == "VariableDeclaration":
                var = TopLevelVariable(scope)
                var_parser = TopLevelVariableSolc(var, top_level_data, self)
                var.set_offset(top_level_data["src"], self._compilation_unit)

                self._compilation_unit.variables_top_level.append(var)
                self._variables_top_level_parser.append(var_parser)
                scope.variables[var.name] = var
                referenceId = top_level_data["id"]
                self.top_level_variables_by_id[referenceId] = var

            elif top_level_data[self.get_key()] == "FunctionDefinition":
                func = FunctionTopLevel(self._compilation_unit, scope)
                scope.functions.add(func)
                func.set_offset(top_level_data["src"], self._compilation_unit)
                func_parser = FunctionSolc(func, top_level_data, None, self)

                self._compilation_unit.functions_top_level.append(func)
                self._functions_top_level_parser.append(func_parser)
                self.add_function_or_modifier_parser(func_parser)

            elif top_level_data[self.get_key()] == "ErrorDefinition":
                custom_error = CustomErrorTopLevel(self._compilation_unit, scope)
                custom_error.set_offset(top_level_data["src"], self._compilation_unit)

                custom_error_parser = CustomErrorSolc(custom_error, top_level_data, None, self)
                scope.custom_errors.add(custom_error)
                self._compilation_unit.custom_errors.append(custom_error)
                self._custom_error_parser.append(custom_error_parser)
                referenceId = top_level_data["id"]
                self.top_level_errors_by_id[referenceId] = custom_error

            elif top_level_data[self.get_key()] == "UserDefinedValueTypeDefinition":
                assert "name" in top_level_data
                alias = top_level_data["name"]
                assert "underlyingType" in top_level_data
                underlying_type = top_level_data["underlyingType"]
                assert (
                    "nodeType" in underlying_type
                    and underlying_type["nodeType"] == "ElementaryTypeName"
                )
                assert "name" in underlying_type

                original_type = ElementaryType(underlying_type["name"])

                type_alias = TypeAliasTopLevel(original_type, alias, scope)
                type_alias.set_offset(top_level_data["src"], self._compilation_unit)
                self._compilation_unit.type_aliases[alias] = type_alias
                scope.type_aliases[alias] = type_alias
                referenceId = top_level_data["id"]
                self.top_level_type_aliases_by_id[referenceId] = type_alias

            elif top_level_data[self.get_key()] == "EventDefinition":
                event = EventTopLevel(scope)
                event.set_offset(top_level_data["src"], self._compilation_unit)

                event_parser = EventTopLevelSolc(event, top_level_data, self)  # type: ignore
                self._events_top_level_parser.append(event_parser)
                scope.events.add(event)
                self._compilation_unit.events_top_level.append(event)
                referenceId = top_level_data["id"]
                self.top_level_events_by_id[referenceId] = event

            else:
                raise SlitherException(f"Top level {top_level_data[self.get_key()]} not supported")

    def _parse_source_unit(self, data: dict, filename: str) -> None:
        if data[self.get_key()] != "SourceUnit":
            return  # handle solc prior 0.3.6

        # match any char for filename
        # filename can contain space, /, -, ..
        name_candidates = re.findall("=+ (.+) =+", filename)
        if name_candidates:
            assert len(name_candidates) == 1
            name: str = name_candidates[0]
        else:
            name = filename

        sourceUnit = -1  # handle old solc, or error
        if "src" in data:
            sourceUnit_candidates = re.findall("[0-9]*:[0-9]*:([0-9]*)", data["src"])
            if len(sourceUnit_candidates) == 1:
                sourceUnit = int(sourceUnit_candidates[0])
        if sourceUnit == -1:
            # if source unit is not found
            # We can still deduce it, by assigning to the last source_code added
            # This works only for crytic compile.
            # which used --combined-json ast, rather than --ast-json
            # As a result -1 is not used as index
            if self._compilation_unit.core.crytic_compile is not None:
                sourceUnit = len(self._compilation_unit.core.source_code)

        self._compilation_unit.source_units[sourceUnit] = name
        if os.path.isfile(name) and name not in self._compilation_unit.core.source_code:
            self._compilation_unit.core.add_source_code(name)
        else:
            lib_name = os.path.join("node_modules", name)
            if os.path.isfile(lib_name) and name not in self._compilation_unit.core.source_code:
                self._compilation_unit.core.add_source_code(lib_name)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Analyze
    ###################################################################################
    ###################################################################################

    @property
    def parsed(self) -> bool:
        return self._parsed

    @property
    def analyzed(self) -> bool:
        return self._analyzed

    def parse_contracts(self) -> None:
        if not self._underlying_contract_to_parser:
            logger.info(
                f"No contracts were found in {self._compilation_unit.core.filename}, check the correct compilation"
            )
        if self._parsed:
            raise Exception("Contract analysis can be run only once!")

        def resolve_remapping_and_renaming(contract_parser: ContractSolc, want: str) -> Contract:
            contract_name = contract_parser.remapping[want]
            target = None
            # For contracts that are imported and aliased e.g. 'import {A as B} from "./C.sol"',
            # we look through the imports's (`Import`) renaming to find the original contract name
            # and then look up the original contract in the import path's scope (`FileScope`).
            for import_ in contract_parser.underlying_contract.file_scope.imports:
                if contract_name in import_.renaming:
                    target = self.compilation_unit.get_scope(
                        import_.filename
                    ).get_contract_from_name(import_.renaming[contract_name])

            # Fallback to the current file scope if the contract is not found in the import path's scope.
            # It is assumed that it isn't possible to defined a contract with the same name as "aliased" names.
            if target is None:
                target = contract_parser.underlying_contract.file_scope.get_contract_from_name(
                    contract_name
                )

            if target == contract_parser.underlying_contract:
                raise InheritanceResolutionError(
                    "Could not resolve contract inheritance. This is likely caused by an import renaming that collides with existing names (see https://github.com/crytic/slither/issues/1758)."
                    f"\n Try changing `contract {target}` ({target.source_mapping}) to a unique name as a workaround."
                    "\n Please share the source code that caused this error here: https://github.com/crytic/slither/issues/"
                )
            assert target, f"Contract {contract_name} not found"
            return target

        # Update of the inheritance
        for contract_parser in self._underlying_contract_to_parser.values():
            ancestors = []
            fathers = []
            father_constructors = []
            missing_inheritance = None

            # Resolve linearized base contracts.
            # Remove the first elem in linearizedBaseContracts as it is the contract itself.
            for i in contract_parser.linearized_base_contracts[1:]:
                if i in contract_parser.remapping:
                    target = resolve_remapping_and_renaming(contract_parser, i)
                    ancestors.append(target)
                elif i in self._contracts_by_id:
                    ancestors.append(self._contracts_by_id[i])
                else:
                    missing_inheritance = i

            # Resolve immediate base contracts and attach references.
            for i, src in contract_parser.baseContracts:
                if i in contract_parser.remapping:
                    target = resolve_remapping_and_renaming(contract_parser, i)
                    fathers.append(target)
                    target.add_reference_from_raw_source(src, self.compilation_unit)
                elif i in self._contracts_by_id:
                    target = self._contracts_by_id[i]
                    fathers.append(target)
                    target.add_reference_from_raw_source(src, self.compilation_unit)
                else:
                    missing_inheritance = i

            # Resolve immediate base constructor calls.
            for i in contract_parser.baseConstructorContractsCalled:
                if i in contract_parser.remapping:
                    target = resolve_remapping_and_renaming(contract_parser, i)
                    father_constructors.append(target)
                elif i in self._contracts_by_id:
                    father_constructors.append(self._contracts_by_id[i])
                else:
                    missing_inheritance = i

            contract_parser.underlying_contract.set_inheritance(
                ancestors, fathers, father_constructors
            )

            if missing_inheritance:
                self._compilation_unit.contracts_with_missing_inheritance.add(
                    contract_parser.underlying_contract
                )
                txt = f"Missing inheritance {contract_parser.underlying_contract} ({contract_parser.compilation_unit.crytic_compile_compilation_unit.unique_id})\n"
                txt += f"Missing inheritance ID: {missing_inheritance}\n"
                if contract_parser.underlying_contract.inheritance:
                    txt += "Inheritance found:\n"
                    for contract_inherited in contract_parser.underlying_contract.inheritance:
                        txt += f"\t - {contract_inherited} (ID {contract_inherited.id})\n"
                contract_parser.log_incorrect_parsing(txt)

                contract_parser.set_is_analyzed(True)
                contract_parser.delete_content()

        contracts_to_be_analyzed = list(self._underlying_contract_to_parser.values())

        # Any contract can refer another contract enum without need for inheritance
        self._analyze_all_enums(contracts_to_be_analyzed)

        [c.set_is_analyzed(False) for c in self._underlying_contract_to_parser.values()]

        libraries = [
            c for c in contracts_to_be_analyzed if c.underlying_contract.contract_kind == "library"
        ]
        contracts_to_be_analyzed = [
            c for c in contracts_to_be_analyzed if c.underlying_contract.contract_kind != "library"
        ]

        # We first parse the struct/variables/functions/contract
        self._analyze_first_part(contracts_to_be_analyzed, libraries)

        [c.set_is_analyzed(False) for c in self._underlying_contract_to_parser.values()]

        # We analyze the struct and parse and analyze the events
        # A contract can refer in the variables a struct or a event from any contract
        # (without inheritance link)
        self._analyze_second_part(contracts_to_be_analyzed, libraries)
        [c.set_is_analyzed(False) for c in self._underlying_contract_to_parser.values()]

        # Then we analyse state variables, functions and modifiers
        self._analyze_third_part(contracts_to_be_analyzed, libraries)
        [c.set_is_analyzed(False) for c in self._underlying_contract_to_parser.values()]

        self._analyze_using_for(contracts_to_be_analyzed, libraries)

        self._parsed = True

    def analyze_contracts(self) -> None:
        if not self._parsed:
            raise SlitherException("Parse the contract before running analyses")
        self._convert_to_slithir()
        if not self._compilation_unit.core.skip_data_dependency:
            compute_dependency(self._compilation_unit)
        self._compilation_unit.compute_storage_layout()
        self._analyzed = True

    def _analyze_all_enums(self, contracts_to_be_analyzed: list[ContractSolc]) -> None:
        while contracts_to_be_analyzed:
            contract = contracts_to_be_analyzed[0]

            contracts_to_be_analyzed = contracts_to_be_analyzed[1:]
            all_father_analyzed = all(
                self._underlying_contract_to_parser[father].is_analyzed
                for father in contract.underlying_contract.inheritance
            )

            if not contract.underlying_contract.inheritance or all_father_analyzed:
                self._analyze_enums(contract)
            else:
                contracts_to_be_analyzed += [contract]

    def _analyze_first_part(
        self,
        contracts_to_be_analyzed: list[ContractSolc],
        libraries: list[ContractSolc],
    ) -> None:
        for lib in libraries:
            self._parse_struct_var_modifiers_functions(lib)

        # Start with the contracts without inheritance
        # Analyze a contract only if all its fathers
        # Were analyzed
        while contracts_to_be_analyzed:
            contract = contracts_to_be_analyzed[0]

            contracts_to_be_analyzed = contracts_to_be_analyzed[1:]
            all_father_analyzed = all(
                self._underlying_contract_to_parser[father].is_analyzed
                for father in contract.underlying_contract.inheritance
            )

            if not contract.underlying_contract.inheritance or all_father_analyzed:
                self._parse_struct_var_modifiers_functions(contract)

            else:
                contracts_to_be_analyzed += [contract]

    def _analyze_second_part(
        self,
        contracts_to_be_analyzed: list[ContractSolc],
        libraries: list[ContractSolc],
    ) -> None:
        for lib in libraries:
            self._analyze_struct_events(lib)

        self._analyze_top_level_variables()
        self._analyze_top_level_structures()
        self._analyze_top_level_events()

        # Start with the contracts without inheritance
        # Analyze a contract only if all its fathers
        # Were analyzed
        while contracts_to_be_analyzed:
            contract = contracts_to_be_analyzed[0]

            contracts_to_be_analyzed = contracts_to_be_analyzed[1:]
            all_father_analyzed = all(
                self._underlying_contract_to_parser[father].is_analyzed
                for father in contract.underlying_contract.inheritance
            )

            if not contract.underlying_contract.inheritance or all_father_analyzed:
                self._analyze_struct_events(contract)

            else:
                contracts_to_be_analyzed += [contract]

    def _analyze_third_part(
        self,
        contracts_to_be_analyzed: list[ContractSolc],
        libraries: list[ContractSolc],
    ) -> None:
        for lib in libraries:
            self._analyze_variables_modifiers_functions(lib)

        # Start with the contracts without inheritance
        # Analyze a contract only if all its fathers
        # Were analyzed
        while contracts_to_be_analyzed:
            contract = contracts_to_be_analyzed[0]

            contracts_to_be_analyzed = contracts_to_be_analyzed[1:]
            all_father_analyzed = all(
                self._underlying_contract_to_parser[father].is_analyzed
                for father in contract.underlying_contract.inheritance
            )

            if not contract.underlying_contract.inheritance or all_father_analyzed:
                self._analyze_variables_modifiers_functions(contract)

            else:
                contracts_to_be_analyzed += [contract]

    def _analyze_using_for(
        self, contracts_to_be_analyzed: list[ContractSolc], libraries: list[ContractSolc]
    ) -> None:
        self._analyze_top_level_using_for()

        for lib in libraries:
            lib.analyze_using_for()

        while contracts_to_be_analyzed:
            contract = contracts_to_be_analyzed[0]

            contracts_to_be_analyzed = contracts_to_be_analyzed[1:]
            all_father_analyzed = all(
                self._underlying_contract_to_parser[father].is_analyzed
                for father in contract.underlying_contract.inheritance
            )

            if not contract.underlying_contract.inheritance or all_father_analyzed:
                contract.analyze_using_for()
                contract.set_is_analyzed(True)
            else:
                contracts_to_be_analyzed += [contract]

    def _analyze_enums(self, contract: ContractSolc) -> None:
        # Enum must be analyzed first
        contract.analyze_enums()
        contract.set_is_analyzed(True)

    def _parse_struct_var_modifiers_functions(self, contract: ContractSolc) -> None:
        contract.parse_structs()  # struct can refer another struct
        contract.parse_state_variables()
        contract.parse_modifiers()
        contract.parse_functions()
        contract.parse_custom_errors()
        contract.parse_type_alias()
        contract.set_is_analyzed(True)

    def _analyze_struct_events(self, contract: ContractSolc) -> None:
        contract.analyze_constant_state_variables()

        # Struct can refer to enum, or state variables
        contract.analyze_structs()
        # Event can refer to struct
        contract.analyze_events()

        contract.analyze_custom_errors()

        # Must be here since it can use top level variables hence they must be analyzed
        contract.analyze_storage_layout()

        contract.set_is_analyzed(True)

    def _analyze_top_level_structures(self) -> None:
        try:
            for struct in self._structures_top_level_parser:
                struct.analyze()
        except (VariableNotFound, KeyError) as e:
            raise SlitherException(f"Missing struct {e} during top level structure analyze") from e

    def _analyze_top_level_variables(self) -> None:
        try:
            for var in self._variables_top_level_parser:
                var.analyze(var)
        except (VariableNotFound, KeyError) as e:
            raise SlitherException(f"Missing {e} during variable analyze") from e

    def _analyze_top_level_events(self) -> None:
        try:
            for event in self._events_top_level_parser:
                event.analyze()
        except (VariableNotFound, KeyError) as e:
            raise SlitherException(f"Missing event {e} during top level event analyze") from e

    def _analyze_params_top_level_function(self) -> None:
        for func_parser in self._functions_top_level_parser:
            func_parser.analyze_params()
            self._compilation_unit.add_function(func_parser.underlying_function)

    def _analyze_top_level_using_for(self) -> None:
        for using_for in self._using_for_top_level_parser:
            using_for.analyze()

    def _analyze_params_custom_error(self) -> None:
        for custom_error_parser in self._custom_error_parser:
            custom_error_parser.analyze_params()

    def _analyze_content_top_level_function(self) -> None:
        try:
            for func_parser in self._functions_top_level_parser:
                func_parser.analyze_content()
        except (VariableNotFound, KeyError) as e:
            raise SlitherException(f"Missing {e} during top level function analyze") from e

    def _analyze_variables_modifiers_functions(self, contract: ContractSolc) -> None:
        # State variables, modifiers and functions can refer to anything

        contract.analyze_params_modifiers()
        contract.analyze_params_functions()
        self._analyze_params_top_level_function()
        self._analyze_params_custom_error()

        contract.analyze_state_variables()

        contract.analyze_content_modifiers()
        contract.analyze_content_functions()
        self._analyze_content_top_level_function()

        contract.set_is_analyzed(True)

    def _convert_to_slithir(self) -> None:
        for contract in self._compilation_unit.contracts:
            contract.add_constructor_variables()

            for func in contract.functions + contract.modifiers:
                try:
                    func.generate_slithir_and_analyze()

                except AttributeError as e:
                    # This can happens for example if there is a call to an interface
                    # And the interface is redefined due to contract's name reuse
                    # But the available version misses some functions
                    self._underlying_contract_to_parser[contract].log_incorrect_parsing(
                        f"Impossible to generate IR for {contract.name}.{func.name} ({func.source_mapping}):\n {e}"
                    )
                except Exception as e:
                    func_expressions = "\n".join([f"\t{ex}" for ex in func.expressions])
                    logger.error(
                        f"\nFailed to generate IR for {contract.name}.{func.name}. Please open an issue https://github.com/crytic/slither/issues.\n{contract.name}.{func.name} ({func.source_mapping}):\n "
                        f"{func_expressions}"
                    )
                    raise e
            try:
                contract.convert_expression_to_slithir_ssa()
            except Exception as e:
                logger.error(
                    f"\nFailed to convert IR to SSA for {contract.name} contract. Please open an issue https://github.com/crytic/slither/issues.\n "
                )
                raise e

        for func in self._compilation_unit.functions_top_level:
            try:
                func.generate_slithir_and_analyze()
            except AttributeError as e:
                logger.error(
                    f"Impossible to generate IR for top level function {func.name} ({func.source_mapping}):\n {e}"
                )
            except Exception as e:
                func_expressions = "\n".join([f"\t{ex}" for ex in func.expressions])
                logger.error(
                    f"\nFailed to generate IR for top level function {func.name}. Please open an issue https://github.com/crytic/slither/issues.\n{func.name} ({func.source_mapping}):\n "
                    f"{func_expressions}"
                )
                raise e

            try:
                func.generate_slithir_ssa({})
            except Exception as e:
                func_expressions = "\n".join([f"\t{ex}" for ex in func.expressions])
                logger.error(
                    f"\nFailed to convert IR to SSA for top level function {func.name}. Please open an issue https://github.com/crytic/slither/issues.\n{func.name} ({func.source_mapping}):\n "
                    f"{func_expressions}"
                )
                raise e

        self._compilation_unit.propagate_function_calls()
        for contract in self._compilation_unit.contracts:
            contract.fix_phi()
            contract.update_read_write_using_ssa()

    # endregion



import logging
from enum import Enum
from typing import Optional, TYPE_CHECKING

from slither.core.expressions.expression import Expression
from slither.core.exceptions import SlitherCoreError

if TYPE_CHECKING:
    from slither.core.solidity_types.type import Type

logger = logging.getLogger("AssignmentOperation")


class AssignmentOperationType(Enum):
    ASSIGN = 0  # =
    ASSIGN_OR = 1  # |=
    ASSIGN_CARET = 2  # ^=
    ASSIGN_AND = 3  # &=
    ASSIGN_LEFT_SHIFT = 4  # <<=
    ASSIGN_RIGHT_SHIFT = 5  # >>=
    ASSIGN_ADDITION = 6  # +=
    ASSIGN_SUBTRACTION = 7  # -=
    ASSIGN_MULTIPLICATION = 8  # *=
    ASSIGN_DIVISION = 9  # /=
    ASSIGN_MODULO = 10  # %=

    @staticmethod
    def get_type(operation_type: str) -> "AssignmentOperationType":
        if operation_type == "=":
            return AssignmentOperationType.ASSIGN
        if operation_type == "|=":
            return AssignmentOperationType.ASSIGN_OR
        if operation_type == "^=":
            return AssignmentOperationType.ASSIGN_CARET
        if operation_type == "&=":
            return AssignmentOperationType.ASSIGN_AND
        if operation_type == "<<=":
            return AssignmentOperationType.ASSIGN_LEFT_SHIFT
        if operation_type == ">>=":
            return AssignmentOperationType.ASSIGN_RIGHT_SHIFT
        if operation_type == "+=":
            return AssignmentOperationType.ASSIGN_ADDITION
        if operation_type == "-=":
            return AssignmentOperationType.ASSIGN_SUBTRACTION
        if operation_type == "*=":
            return AssignmentOperationType.ASSIGN_MULTIPLICATION
        if operation_type == "/=":
            return AssignmentOperationType.ASSIGN_DIVISION
        if operation_type == "%=":
            return AssignmentOperationType.ASSIGN_MODULO

        raise SlitherCoreError(f"get_type: Unknown operation type {operation_type})")

    def __str__(self) -> str:
        if self == AssignmentOperationType.ASSIGN:
            return "="
        if self == AssignmentOperationType.ASSIGN_OR:
            return "|="
        if self == AssignmentOperationType.ASSIGN_CARET:
            return "^="
        if self == AssignmentOperationType.ASSIGN_AND:
            return "&="
        if self == AssignmentOperationType.ASSIGN_LEFT_SHIFT:
            return "<<="
        if self == AssignmentOperationType.ASSIGN_RIGHT_SHIFT:
            return ">>="
        if self == AssignmentOperationType.ASSIGN_ADDITION:
            return "+="
        if self == AssignmentOperationType.ASSIGN_SUBTRACTION:
            return "-="
        if self == AssignmentOperationType.ASSIGN_MULTIPLICATION:
            return "*="
        if self == AssignmentOperationType.ASSIGN_DIVISION:
            return "/="
        if self == AssignmentOperationType.ASSIGN_MODULO:
            return "%="
        raise SlitherCoreError(f"str: Unknown operation type {self})")


class AssignmentOperation(Expression):
    def __init__(
        self,
        left_expression: Expression,
        right_expression: Expression,
        expression_type: AssignmentOperationType,
        expression_return_type: Optional["Type"],
    ) -> None:
        assert isinstance(left_expression, Expression)
        assert isinstance(right_expression, Expression)
        super().__init__()
        left_expression.set_lvalue()
        self._expressions = [left_expression, right_expression]
        self._type: AssignmentOperationType = expression_type
        self._expression_return_type: Type | None = expression_return_type

    @property
    def expressions(self) -> list[Expression]:
        return self._expressions

    @property
    def expression_return_type(self) -> Optional["Type"]:
        return self._expression_return_type

    @property
    def expression_left(self) -> Expression:
        return self._expressions[0]

    @property
    def expression_right(self) -> Expression:
        return self._expressions[1]

    @property
    def type(self) -> Optional["AssignmentOperationType"]:
        return self._type

    def __str__(self) -> str:
        return str(self.expression_left) + " " + str(self.type) + " " + str(self.expression_right)



"""
Module detecting usage of inline assembly
"""

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class Assembly(AbstractDetector):
    """
    Detect usage of inline assembly
    """

    ARGUMENT = "assembly"
    HELP = "Assembly usage"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#assembly-usage"

    WIKI_TITLE = "Assembly usage"
    WIKI_DESCRIPTION = "The use of assembly is error-prone and should be avoided."
    WIKI_RECOMMENDATION = "Do not use `evm` assembly."

    @staticmethod
    def _contains_inline_assembly_use(node: Node) -> bool:
        """
             Check if the node contains ASSEMBLY type
        Returns:
            (bool)
        """
        return node.type == NodeType.ASSEMBLY

    def detect_assembly(self, contract: Contract) -> list[tuple[FunctionContract, list[Node]]]:
        ret = []
        for f in contract.functions:
            if f.contract_declarer != contract:
                continue
            nodes = f.nodes
            assembly_nodes = [n for n in nodes if self._contains_inline_assembly_use(n)]
            if assembly_nodes:
                ret.append((f, assembly_nodes))
        return ret

    def _detect(self) -> list[Output]:
        """Detect the functions that use inline assembly"""
        results = []
        for c in self.contracts:
            values = self.detect_assembly(c)
            for func, nodes in values:
                info: DETECTOR_INFO = [func, " uses assembly\n"]

                # sort the nodes to get deterministic results
                nodes.sort(key=lambda x: x.node_id)

                for node in nodes:
                    info += ["\t- ", node, "\n"]

                res = self.generate_result(info)
                results.append(res)

        return results



from pathlib import Path

from slither import Slither


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
NAME_RESOLUTION_TEST_ROOT = Path(TEST_DATA_DIR, "name_resolution")


def _sort_references_lines(refs: list) -> list:
    return sorted([ref.lines[0] for ref in refs])


def test_name_resolution_compact(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(
        Path(NAME_RESOLUTION_TEST_ROOT, "shadowing_compact.sol").as_posix(), solc=solc_path
    )
    contract = slither.get_contract_from_name("B")[0]
    x = contract.get_state_variable_from_name("x")
    assert _sort_references_lines(x.references) == [5]


def test_name_resolution_legacy_post_0_5_0(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.5.0")
    slither = Slither(
        Path(NAME_RESOLUTION_TEST_ROOT, "shadowing_legacy_post_0_5_0.sol").as_posix(),
        solc=solc_path,
    )
    contract = slither.get_contract_from_name("B")[0]
    x = contract.get_state_variable_from_name("x")
    assert _sort_references_lines(x.references) == [5]


def test_name_resolution_legacy_pre_0_5_0(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.4.12")
    slither = Slither(
        Path(NAME_RESOLUTION_TEST_ROOT, "shadowing_legacy_pre_0_5_0.sol").as_posix(),
        solc=solc_path,
        force_legacy=True,
    )
    contract = slither.get_contract_from_name("B")[0]
    function = contract.get_function_from_signature("a()")
    x = function.get_local_variable_from_name("x")
    assert _sort_references_lines(x.references) == [5]



import sys
from slither.slither import Slither

if len(sys.argv) != 2:
    print("python function_writing.py functions_writing.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

# Get the contract
contracts = slither.get_contract_from_name("Contract")
assert len(contracts) == 1
contract = contracts[0]
# Get the variable
var_a = contract.get_state_variable_from_name("a")

# Get the functions writing the variable
functions_writing_a = contract.get_functions_writing_to_variable(var_a)

# Print the result
print(f'The function writing "a" are {[f.name for f in functions_writing_a]}')



"""
Module detecting suicidal contract

A suicidal contract is an unprotected function that calls selfdestruct
"""

from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class Suicidal(AbstractDetector):
    """
    Unprotected function detector
    """

    ARGUMENT = "suicidal"
    HELP = "Functions allowing anyone to destruct the contract"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#suicidal"

    WIKI_TITLE = "Suicidal"
    WIKI_DESCRIPTION = "Unprotected call to a function executing `selfdestruct`/`suicide`."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Suicidal{
    function kill() public{
        selfdestruct(msg.sender);
    }
}
```
Bob calls `kill` and destructs the contract."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Protect access to all sensitive functions."

    @staticmethod
    def detect_suicidal_func(func: FunctionContract) -> bool:
        """Detect if the function is suicidal

        Detect the public functions calling suicide/selfdestruct without protection
        Returns:
            (bool): True if the function is suicidal
        """

        if func.is_constructor:
            return False

        if func.visibility not in ["public", "external"]:
            return False

        calls = [ir.function.name for ir in func.all_internal_calls()]
        if not ("suicide(address)" in calls or "selfdestruct(address)" in calls):
            return False

        if func.is_protected():
            return False

        return True

    def detect_suicidal(self, contract: Contract) -> list[FunctionContract]:
        ret = []
        for f in contract.functions_declared:
            if self.detect_suicidal_func(f):
                ret.append(f)
        return ret

    def _detect(self) -> list[Output]:
        """Detect the suicidal functions"""
        results = []
        for c in self.contracts:
            functions = self.detect_suicidal(c)
            for func in functions:
                info: DETECTOR_INFO = [func, " allows anyone to destruct the contract\n"]

                res = self.generate_result(info)

                results.append(res)

        return results



#


from slither.core.solidity_types import ElementaryType
from slither.slithir.operations import (
    Phi,
    InternalCall,
)
from slither.slithir.variables import (
    Constant,
)


def test_interface_conversion_and_call_resolution(slither_from_vyper_source):
    with slither_from_vyper_source(
        """
interface Test:
    def foo() -> (int128, uint256): nonpayable

@internal
def foo() -> (int128, int128):
    return 2, 3

@external
def bar():
    a: int128 = 0
    b: int128 = 0
    (a, b) = self.foo()

    x: address = 0x0000000000000000000000000000000000000000
    c: uint256 = 0
    a, c = Test(x).foo()
"""
    ) as sl:
        interface = next(iter(x for x in sl.contracts if x.is_interface))
        contract = next(iter(x for x in sl.contracts if not x.is_interface))
        func = contract.get_function_from_signature("bar()")
        (contract, ir) = func.high_level_calls[0]
        assert contract == interface
        assert ir.function.signature_str == "foo() returns(int128,uint256)"


def test_phi_entry_point_internal_call(slither_from_vyper_source):
    with slither_from_vyper_source(
        """
counter: uint256
@internal
def b(y: uint256):
    self.counter = y

@external
def a(x: uint256):
    self.b(x)
    self.b(1)
"""
    ) as sl:
        f = sl.contracts[0].get_function_from_signature("b(uint256)")
        assert (
            len(
                [
                    ssanode
                    for node in f.nodes
                    for ssanode in node.irs_ssa
                    if isinstance(ssanode, Phi)
                ]
            )
            == 1
        )


def test_call_with_default_args(slither_from_vyper_source):
    with slither_from_vyper_source(
        """
counter: uint256
@internal
def c(y: uint256, config: bool = True):
    self.counter = y
@external
def a(x: uint256):
    self.c(x)
    self.c(1)
@external
def b(x: uint256):
    self.c(x, False)
    self.c(1, False)
"""
    ) as sl:
        a = sl.contracts[0].get_function_from_signature("a(uint256)")
        for node in a.nodes:
            for op in node.irs_ssa:
                if isinstance(op, InternalCall) and op.function.name == "c":
                    assert len(op.arguments) == 2
                    assert op.arguments[1] == Constant("True", ElementaryType("bool"))
        b = sl.contracts[0].get_function_from_signature("b(uint256)")
        for node in b.nodes:
            for op in node.irs_ssa:
                if isinstance(op, InternalCall) and op.function.name == "c":
                    assert len(op.arguments) == 2
                    assert op.arguments[1] == Constant("False", ElementaryType("bool"))



import logging
from pathlib import Path

from slither.core.declarations import Contract
from slither.tools.properties.addresses.address import Addresses
from slither.tools.properties.utils import write_file

logger = logging.getLogger("Slither")


def generate_solidity_properties(
    contract: Contract, type_property: str, solidity_properties: str, output_dir: Path
) -> Path:
    solidity_import = 'import "./interfaces.sol";\n'
    solidity_import += f'import "../{contract.source_mapping.filename.short}";'

    test_contract_name = f"Properties{contract.name}{type_property}"

    solidity_content = (
        f"{solidity_import}\ncontract {test_contract_name} is CryticInterface,{contract.name}"
    )
    solidity_content += f"{{\n\n{solidity_properties}\n}}\n"

    filename = f"{test_contract_name}.sol"
    write_file(output_dir, filename, solidity_content)

    return Path(filename)


def generate_test_contract(
    contract: Contract,
    type_property: str,
    output_dir: Path,
    property_file: Path,
    initialization_recommendation: str,
) -> tuple[str, str]:
    test_contract_name = f"Test{contract.name}{type_property}"
    properties_name = f"Properties{contract.name}{type_property}"

    content = ""
    content += f'import "./{property_file}";\n'
    content += f"contract {test_contract_name} is {properties_name} {{\n"
    content += "\tconstructor() public{\n"
    content += "\t\t// Existing addresses:\n"
    content += "\t\t// - crytic_owner: If the contract has an owner, it must be crytic_owner\n"
    content += "\t\t// - crytic_user: Legitimate user\n"
    content += "\t\t// - crytic_attacker: Attacker\n"
    content += "\t\t// \n"
    content += initialization_recommendation
    content += "\t\t// \n"
    content += "\t\t// \n"
    content += "\t\t// Update the following if totalSupply and balanceOf are external functions or state variables:\n\n"
    content += "\t\tinitialTotalSupply = totalSupply();\n"
    content += "\t\tinitialBalance_owner = balanceOf(crytic_owner);\n"
    content += "\t\tinitialBalance_user = balanceOf(crytic_user);\n"
    content += "\t\tinitialBalance_attacker = balanceOf(crytic_attacker);\n"

    content += "\t}\n}\n"

    filename = f"{test_contract_name}.sol"
    write_file(output_dir, filename, content, allow_overwrite=False)

    return filename, test_contract_name


def generate_solidity_interface(output_dir: Path, addresses: Addresses):
    content = f"""
contract CryticInterface{{
    address internal crytic_owner = address({addresses.owner});
    address internal crytic_user = address({addresses.user});
    address internal crytic_attacker = address({addresses.attacker});
    uint internal initialTotalSupply;
    uint internal initialBalance_owner;
    uint internal initialBalance_user;
    uint internal initialBalance_attacker;
}}"""

    # Static file, we discard if it exists as it should never change
    write_file(output_dir, "interfaces.sol", content, discard_if_exist=True)



"""
Module detecting local variable shadowing
"""

from slither.core.declarations.contract import Contract
from slither.core.declarations.event import Event
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.modifier import Modifier
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class LocalShadowing(AbstractDetector):
    """
    Local variable shadowing
    """

    ARGUMENT = "shadowing-local"
    HELP = "Local variables shadowing"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#local-variable-shadowing"

    WIKI_TITLE = "Local variable shadowing"
    WIKI_DESCRIPTION = "Detection of shadowing using local variables."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
pragma solidity ^0.4.24;

contract Bug {
    uint owner;

    function sensitive_function(address owner) public {
        // ...
        require(owner == msg.sender);
    }

    function alternate_sensitive_function() public {
        address owner = msg.sender;
        // ...
        require(owner == msg.sender);
    }
}
```
`sensitive_function.owner` shadows `Bug.owner`. As a result, the use of `owner` in `sensitive_function` might be incorrect."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Rename the local variables that shadow another component."

    OVERSHADOWED_FUNCTION = "function"
    OVERSHADOWED_MODIFIER = "modifier"
    OVERSHADOWED_STATE_VARIABLE = "state variable"
    OVERSHADOWED_EVENT = "event"
    OVERSHADOWED_RETURN_VARIABLE = "return variable"

    def detect_shadowing_definitions(
        self, contract: Contract
    ) -> list[
        tuple[LocalVariable, list[tuple[str, StateVariable]]]
        | tuple[LocalVariable, list[tuple[str, FunctionContract]]]
        | tuple[LocalVariable, list[tuple[str, Modifier]]]
        | tuple[LocalVariable, list[tuple[str, Event]]]
    ]:
        """Detects if functions, access modifiers, events, state variables, and local variables are named after
        reserved keywords. Any such definitions are returned in a list.

        Returns:
            list of tuple: (type, contract name, definition)"""
        result: list[
            tuple[LocalVariable, list[tuple[str, StateVariable]]]
            | tuple[LocalVariable, list[tuple[str, FunctionContract]]]
            | tuple[LocalVariable, list[tuple[str, Modifier]]]
            | tuple[LocalVariable, list[tuple[str, Event]]]
        ] = []

        # Loop through all functions + modifiers in this contract.
        for function in contract.functions + list(contract.modifiers):
            # We should only look for functions declared directly in this contract (not in a base contract).
            if function.contract_declarer != contract:
                continue

            # This function was declared in this contract, we check what its local variables might shadow.
            for variable in function.variables:
                overshadowed = []
                for scope_contract in [contract] + contract.inheritance:
                    # Check functions
                    for scope_function in scope_contract.functions_declared:
                        if variable.name == scope_function.name:
                            overshadowed.append((self.OVERSHADOWED_FUNCTION, scope_function))
                    # Check modifiers
                    for scope_modifier in scope_contract.modifiers_declared:
                        if variable.name == scope_modifier.name:
                            overshadowed.append((self.OVERSHADOWED_MODIFIER, scope_modifier))
                    # Check events
                    for scope_event in scope_contract.events_declared:
                        if variable.name == scope_event.name:
                            overshadowed.append((self.OVERSHADOWED_EVENT, scope_event))
                    # Check state variables
                    for scope_state_variable in scope_contract.state_variables_declared:
                        if variable.name == scope_state_variable.name:
                            overshadowed.append(
                                (self.OVERSHADOWED_STATE_VARIABLE, scope_state_variable)
                            )
                    # Check named return variables
                    for named_return in function.returns:
                        # Shadowed local delcarations in the same function will have "_scope_" in their name.
                        # See `FunctionSolc._add_local_variable`
                        if (
                            "_scope_" in variable.name
                            and variable.name.split("_scope_")[0] == named_return.name
                        ):
                            overshadowed.append((self.OVERSHADOWED_RETURN_VARIABLE, named_return))

                # If we have found any overshadowed objects, we'll want to add it to our result list.
                if overshadowed:
                    result.append((variable, overshadowed))

        return result

    def _detect(self) -> list[Output]:
        """Detect shadowing local variables

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'shadow'}

        """

        results = []
        for contract in self.contracts:
            shadows = self.detect_shadowing_definitions(contract)
            if shadows:
                for shadow in shadows:
                    local_variable = shadow[0]
                    overshadowed = shadow[1]
                    info: DETECTOR_INFO = [local_variable, " shadows:\n"]
                    for overshadowed_entry in overshadowed:
                        info += [
                            "\t- ",
                            overshadowed_entry[1],
                            f" ({overshadowed_entry[0]})\n",
                        ]

                    # Generate relevant JSON data for this shadowing definition.
                    res = self.generate_result(info)

                    results.append(res)

        return results



"""
Module detecting dangerous use of block.timestamp

"""

from slither.analyses.data_dependency.data_dependency import is_dependent
from slither.core.cfg.node import Node
from slither.core.declarations import Function, Contract, FunctionContract
from slither.core.declarations.solidity_variables import (
    SolidityVariableComposed,
    SolidityVariable,
)
from slither.core.variables import Variable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Binary, BinaryType
from slither.utils.output import Output


def _timestamp(func: Function) -> list[Node]:
    ret = set()
    for node in func.nodes:
        if node.contains_require_or_assert():
            for var in node.variables_read:
                if is_dependent(var, SolidityVariableComposed("block.timestamp"), node):
                    ret.add(node)
                if is_dependent(var, SolidityVariable("now"), node):
                    ret.add(node)
        for ir in node.irs:
            if isinstance(ir, Binary) and BinaryType.return_bool(ir.type):
                for var_read in ir.read:
                    if not isinstance(var_read, (Variable, SolidityVariable)):
                        continue
                    if is_dependent(var_read, SolidityVariableComposed("block.timestamp"), node):
                        ret.add(node)
                    if is_dependent(var_read, SolidityVariable("now"), node):
                        ret.add(node)
    return sorted(ret, key=lambda x: x.node_id)


def _detect_dangerous_timestamp(
    contract: Contract,
) -> list[tuple[FunctionContract, list[Node]]]:
    """
    Args:
        contract (Contract)
    Returns:
        list((Function), (list (Node)))
    """
    ret = []
    for f in [f for f in contract.functions if f.contract_declarer == contract]:
        nodes: list[Node] = _timestamp(f)
        if nodes:
            ret.append((f, nodes))
    return ret


class Timestamp(AbstractDetector):
    ARGUMENT = "timestamp"
    HELP = "Dangerous usage of `block.timestamp`"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#block-timestamp"

    WIKI_TITLE = "Block timestamp"
    WIKI_DESCRIPTION = (
        "Dangerous usage of `block.timestamp`. `block.timestamp` can be manipulated by miners."
    )
    WIKI_EXPLOIT_SCENARIO = """"Bob's contract relies on `block.timestamp` for its randomness. Eve is a miner and manipulates `block.timestamp` to exploit Bob's contract."""
    WIKI_RECOMMENDATION = "Avoid relying on `block.timestamp`."

    def _detect(self) -> list[Output]:
        """"""
        results = []

        for c in self.contracts:
            dangerous_timestamp = _detect_dangerous_timestamp(c)
            for func, nodes in dangerous_timestamp:
                info: DETECTOR_INFO = [func, " uses timestamp for comparisons\n"]

                info += ["\tDangerous comparisons:\n"]

                # sort the nodes to get deterministic results
                nodes.sort(key=lambda x: x.node_id)

                for node in nodes:
                    info += ["\t- ", node, "\n"]

                res = self.generate_result(info)

                results.append(res)

        return results



import logging
from typing import Optional, Union, TYPE_CHECKING

from slither.core.cfg.node import NodeType, link_nodes, insert_node, Node
from slither.core.cfg.scope import Scope
from slither.core.declarations.contract import Contract
from slither.core.declarations.function import (
    Function,
    ModifierStatements,
    FunctionType,
)
from slither.core.declarations.modifier import Modifier
from slither.core.declarations.function_contract import FunctionContract
from slither.core.expressions import AssignmentOperation
from slither.core.source_mapping.source_mapping import Source
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.local_variable_init_from_tuple import LocalVariableInitFromTuple
from slither.solc_parsing.cfg.node import NodeSolc
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.exceptions import ParsingError
from slither.solc_parsing.expressions.expression_parsing import parse_expression
from slither.solc_parsing.variables.local_variable import LocalVariableSolc
from slither.solc_parsing.variables.local_variable_init_from_tuple import (
    LocalVariableInitFromTupleSolc,
)
from slither.solc_parsing.variables.variable_declaration import MultipleVariablesDeclaration
from slither.utils.expression_manipulations import SplitTernaryExpression
from slither.visitors.expression.export_values import ExportValues
from slither.visitors.expression.has_conditional import HasConditional
from slither.solc_parsing.yul.parse_yul import YulBlock

if TYPE_CHECKING:
    from slither.core.expressions.expression import Expression
    from slither.solc_parsing.declarations.contract import ContractSolc
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
    from slither.core.compilation_unit import SlitherCompilationUnit


LOGGER = logging.getLogger("FunctionSolc")


def link_underlying_nodes(node1: NodeSolc, node2: NodeSolc):
    link_nodes(node1.underlying_node, node2.underlying_node)


class FunctionSolc(CallerContextExpression):
    # elems = [(type, name)]

    def __init__(
        self,
        function: Function,
        function_data: dict,
        contract_parser: Optional["ContractSolc"],
        slither_parser: "SlitherCompilationUnitSolc",
    ) -> None:
        self._slither_parser: SlitherCompilationUnitSolc = slither_parser
        self._contract_parser = contract_parser
        self._function = function

        # Only present if compact AST
        if self.is_compact_ast:
            self._function.name = function_data["name"]
        else:
            self._function.name = function_data["attributes"][self.get_key()]

        if "id" in function_data:
            self._function.id = function_data["id"]

        self._functionNotParsed = function_data
        self._returnsNotParsed: list[dict] = []
        self._params_was_analyzed = False
        self._content_was_analyzed = False

        self._counter_scope_local_variables = 0
        # variable renamed will map the solc id
        # to the variable. It only works for compact format
        # Later if an expression provides the referencedDeclaration attr
        # we can retrieve the variable
        # It only matters if two variables have the same name in the function
        # which is only possible with solc > 0.5
        self._variables_renamed: dict[int, LocalVariableSolc | LocalVariableInitFromTupleSolc] = {}

        self._analyze_type()

        self._node_to_nodesolc: dict[Node, NodeSolc] = {}
        self._node_to_yulobject: dict[Node, YulBlock] = {}

        self._local_variables_parser: list[LocalVariableSolc | LocalVariableInitFromTupleSolc] = []

        if "documentation" in function_data:
            function.has_documentation = True

    @property
    def underlying_function(self) -> Function:
        return self._function

    @property
    def contract_parser(self) -> Optional["ContractSolc"]:
        return self._contract_parser

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self._function.compilation_unit

    ###################################################################################
    ###################################################################################
    # region AST format
    ###################################################################################
    ###################################################################################

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    def get_children(self, key: str) -> str:
        if self.is_compact_ast:
            return key
        return "children"

    @property
    def is_compact_ast(self):
        return self._slither_parser.is_compact_ast

    # endregion
    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################

    @property
    def variables_renamed(
        self,
    ) -> dict[int, LocalVariableSolc | LocalVariableInitFromTupleSolc]:
        return self._variables_renamed

    def _add_local_variable(
        self, local_var_parser: LocalVariableSolc | LocalVariableInitFromTupleSolc
    ) -> None:
        # If two local variables have the same name
        # We add a suffix to the new variable
        # This is done to prevent collision during SSA translation
        # Use of while in case of collision
        # In the worst case, the name will be really long
        if local_var_parser.underlying_variable.name:
            known_variables = [v.name for v in self._function.variables]
            while local_var_parser.underlying_variable.name in known_variables:
                local_var_parser.underlying_variable.name += (
                    f"_scope_{self._counter_scope_local_variables}"
                )
                self._counter_scope_local_variables += 1
                known_variables = [v.name for v in self._function.variables]
        if local_var_parser.reference_id is not None:
            self._variables_renamed[local_var_parser.reference_id] = local_var_parser
        self._function.variables_as_dict[local_var_parser.underlying_variable.name] = (
            local_var_parser.underlying_variable
        )
        self._local_variables_parser.append(local_var_parser)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Analyses
    ###################################################################################
    ###################################################################################

    @property
    def function_not_parsed(self) -> dict:
        return self._functionNotParsed

    def _analyze_type(self) -> None:
        """
        Analyz the type of the function
        Myst be called in the constructor as the name might change according to the function's type
        For example both the fallback and the receiver will have an empty name
        :return:
        """
        if self.is_compact_ast:
            attributes = self._functionNotParsed
        else:
            attributes = self._functionNotParsed["attributes"]

        if self._function.name == "":
            self._function.function_type = FunctionType.FALLBACK
            # 0.6.x introduced the receiver function
            # It has also an empty name, so we need to check the kind attribute
            if "kind" in attributes:
                if attributes["kind"] == "receive":
                    self._function.function_type = FunctionType.RECEIVE
        else:
            self._function.function_type = FunctionType.NORMAL

        if isinstance(self._function, FunctionContract):
            if self._function.name == self._function.contract_declarer.name:
                self._function.function_type = FunctionType.CONSTRUCTOR

    def _analyze_attributes(self) -> None:
        if self.is_compact_ast:
            attributes = self._functionNotParsed
        else:
            attributes = self._functionNotParsed["attributes"]

        if "stateMutability" in attributes:
            if attributes["stateMutability"] == "payable":
                self._function.payable = True
            elif attributes["stateMutability"] == "pure":
                self._function.pure = True
                self._function.view = True
            elif attributes["stateMutability"] == "view":
                self._function.view = True

        if "constant" in attributes:
            self._function.view = attributes["constant"]

        if attributes.get("isConstructor"):
            self._function.function_type = FunctionType.CONSTRUCTOR

        if "kind" in attributes:
            if attributes["kind"] == "constructor":
                self._function.function_type = FunctionType.CONSTRUCTOR

        if "visibility" in attributes:
            self._function.visibility = attributes["visibility"]
        # old solc
        elif "public" in attributes:
            if attributes["public"]:
                self._function.visibility = "public"
            else:
                self._function.visibility = "private"
        else:
            self._function.visibility = "public"

        if "payable" in attributes:
            self._function.payable = attributes["payable"]

        if "baseFunctions" in attributes:
            overrides_ids = attributes["baseFunctions"]
            if len(overrides_ids) > 0:
                for f_id in overrides_ids:
                    funcs = self.slither_parser.functions_by_id[f_id]
                    for f in funcs:
                        # Do not consider leaf contracts as overrides.
                        # B is A { function a() override {} } and C is A { function a() override {} } override A.a(), not each other.
                        if (
                            f.contract == self._function.contract
                            or f.contract in self._function.contract.inheritance
                        ):
                            self._function.overrides.append(f)
                            f.overridden_by.append(self._function)

        # Attaches reference to override specifier e.g. X is referenced by `function a() override(X)`
        if "overrides" in attributes and isinstance(attributes["overrides"], dict):
            for override in attributes["overrides"].get("overrides", []):
                refId = override["referencedDeclaration"]
                overridden_contract = self.slither_parser.contracts_by_id.get(refId, None)
                if overridden_contract:
                    overridden_contract.add_reference_from_raw_source(
                        override["src"], self.compilation_unit
                    )

        if "virtual" in attributes:
            self._function.is_virtual = attributes["virtual"]

    def analyze_params(self) -> None:
        # Can be re-analyzed due to inheritance
        if self._params_was_analyzed:
            return

        self._params_was_analyzed = True

        self._analyze_attributes()

        if self.is_compact_ast:
            params = self._functionNotParsed["parameters"]
            returns = self._functionNotParsed["returnParameters"]
        else:
            children = self._functionNotParsed[self.get_children("children")]
            # It uses to be
            # params = children[0]
            # returns = children[1]
            # But from Solidity 0.6.3 to 0.6.10 (included)
            # Comment above a function might be added in the children
            child_iter = iter(
                [child for child in children if child[self.get_key()] == "ParameterList"]
            )
            params = next(child_iter)
            returns = next(child_iter)

        if params:
            self._parse_params(params)
        if returns:
            self._parse_returns(returns)

    def analyze_content(self) -> None:
        if self._content_was_analyzed:
            return

        self._content_was_analyzed = True

        if self.is_compact_ast:
            body = self._functionNotParsed.get("body", None)
            return_params = self._functionNotParsed.get("returnParameters", None)

            if body and body[self.get_key()] == "Block":
                self._function.is_implemented = True
                self._parse_cfg(body)

            for modifier in self._functionNotParsed["modifiers"]:
                self._parse_modifier(modifier)

        else:
            children = self._functionNotParsed[self.get_children("children")]
            return_params = children[1]
            self._function.is_implemented = False
            for child in children[2:]:
                if child[self.get_key()] == "Block":
                    self._function.is_implemented = True
                    self._parse_cfg(child)

            # Parse modifier after parsing all the block
            # In the case a local variable is used in the modifier
            for child in children[2:]:
                if child[self.get_key()] == "ModifierInvocation":
                    self._parse_modifier(child)

        for local_var_parser in self._local_variables_parser:
            local_var_parser.analyze(self)

        for node_parser in self._node_to_nodesolc.values():
            node_parser.analyze_expressions(self)

        for yul_parser in self._node_to_yulobject.values():
            yul_parser.analyze_expressions()

        self._rewrite_ternary_as_if_else()

        self._remove_alone_endif()

        if return_params:
            self._fix_implicit_return(return_params)

        if self._function.entry_point:
            self._update_reachability(self._function.entry_point)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Nodes
    ###################################################################################
    ###################################################################################

    def _new_node(
        self, node_type: NodeType, src: str | Source, scope: Union[Scope, "Function"]
    ) -> NodeSolc:
        node = self._function.new_node(node_type, src, scope)
        node_parser = NodeSolc(node)
        self._node_to_nodesolc[node] = node_parser
        return node_parser

    def _new_yul_block(self, src: str | dict, father_scope: Scope | Function) -> YulBlock:
        scope = Scope(False, True, father_scope)
        node = self._function.new_node(NodeType.ASSEMBLY, src, scope)
        contract = None
        if isinstance(self._function, FunctionContract):
            contract = self._function.contract
        yul_object = YulBlock(
            contract,
            node,
            [self._function.name, f"asm_{len(self._node_to_yulobject)}"],
            scope,
        )
        self._node_to_yulobject[node] = yul_object
        return yul_object

    # endregion
    ###################################################################################
    ###################################################################################
    # region Parsing function
    ###################################################################################
    ###################################################################################

    def _parse_if(self, if_statement: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        # IfStatement = 'if' '(' Expression ')' Statement ( 'else' Statement )?
        falseStatement = None

        if self.is_compact_ast:
            condition = if_statement["condition"]
            # Note: check if the expression could be directly
            # parsed here
            condition_node = self._new_node(NodeType.IF, condition["src"], scope)
            condition_node.add_unparsed_expression(condition)
            link_underlying_nodes(node, condition_node)
            true_scope = Scope(scope.is_checked, False, scope)
            trueStatement = self._parse_statement(
                if_statement["trueBody"], condition_node, true_scope
            )
            if if_statement.get("falseBody"):
                false_scope = Scope(scope.is_checked, False, scope)
                falseStatement = self._parse_statement(
                    if_statement["falseBody"], condition_node, false_scope
                )
        else:
            children = if_statement[self.get_children("children")]
            condition = children[0]
            # Note: check if the expression could be directly
            # parsed here
            condition_node = self._new_node(NodeType.IF, condition["src"], scope)
            condition_node.add_unparsed_expression(condition)
            link_underlying_nodes(node, condition_node)
            true_scope = Scope(scope.is_checked, False, scope)
            trueStatement = self._parse_statement(children[1], condition_node, true_scope)
            if len(children) == 3:
                false_scope = Scope(scope.is_checked, False, scope)
                falseStatement = self._parse_statement(children[2], condition_node, false_scope)

        endIf_node = self._new_node(NodeType.ENDIF, if_statement["src"], scope)
        link_underlying_nodes(trueStatement, endIf_node)

        if falseStatement:
            link_underlying_nodes(falseStatement, endIf_node)
        else:
            link_underlying_nodes(condition_node, endIf_node)
        return endIf_node

    def _parse_while(self, whilte_statement: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        # WhileStatement = 'while' '(' Expression ')' Statement

        node_startWhile = self._new_node(NodeType.STARTLOOP, whilte_statement["src"], scope)

        body_scope = Scope(scope.is_checked, False, scope)
        if self.is_compact_ast:
            node_condition = self._new_node(
                NodeType.IFLOOP, whilte_statement["condition"]["src"], scope
            )
            node_condition.add_unparsed_expression(whilte_statement["condition"])
            statement = self._parse_statement(whilte_statement["body"], node_condition, body_scope)
        else:
            children = whilte_statement[self.get_children("children")]
            expression = children[0]
            node_condition = self._new_node(NodeType.IFLOOP, expression["src"], scope)
            node_condition.add_unparsed_expression(expression)
            statement = self._parse_statement(children[1], node_condition, body_scope)

        node_endWhile = self._new_node(NodeType.ENDLOOP, whilte_statement["src"], scope)

        link_underlying_nodes(node, node_startWhile)
        link_underlying_nodes(node_startWhile, node_condition)
        link_underlying_nodes(statement, node_condition)
        link_underlying_nodes(node_condition, node_endWhile)

        return node_endWhile

    def _parse_for_compact_ast(
        self, statement: dict
    ) -> tuple[dict | None, dict | None, dict | None, dict]:
        body = statement["body"]
        init_expression = statement.get("initializationExpression")
        condition = statement.get("condition")
        loop_expression = statement.get("loopExpression")

        return init_expression, condition, loop_expression, body

    def _parse_for_legacy_ast(
        self, statement: dict
    ) -> tuple[dict | None, dict | None, dict | None, dict]:
        # if we're using an old version of solc (anything below and including 0.4.11) or if the user
        # explicitly enabled compact ast, we might need to make some best-effort guesses
        children = statement[self.get_children("children")]

        # there should always be at least one, and never more than 4, children
        assert 1 <= len(children) <= 4

        # the last element of the children array must be the body, since it's mandatory
        # however, it might be a single expression
        body = children[-1]

        if len(children) == 4:
            # handle the first trivial case - if there are four children we know exactly what they are
            pre, cond, post = children[0], children[1], children[2]
        elif len(children) == 1:
            # handle the second trivial case - if there is only one child we know there are no expressions
            pre, cond, post = None, None, None
        else:
            attributes = statement.get("attributes")

            def has_hint(key):
                return key in attributes and not attributes[key]

            if attributes and any(
                map(
                    has_hint,
                    ["condition", "initializationExpression", "loopExpression"],
                )
            ):
                # if we have attribute hints, rely on those

                if len(children) == 2:
                    # we're missing two expressions, find the one we have
                    if not has_hint("initializationExpression"):
                        pre, cond, post = children[0], None, None
                    elif not has_hint("condition"):
                        pre, cond, post = None, children[0], None
                    else:  # if not has_hint('loopExpression'):
                        pre, cond, post = None, None, children[0]
                else:  # len(children) == 3
                    # we're missing one expression, figure out what it is
                    if has_hint("initializationExpression"):
                        pre, cond, post = None, children[0], children[1]
                    elif has_hint("condition"):
                        pre, cond, post = children[0], None, children[1]
                    else:  # if has_hint('loopExpression'):
                        pre, cond, post = children[0], children[1], None
            else:
                # we don't have attribute hints, and it's impossible to be 100% accurate here
                # let's just try our best

                first_type = children[0][self.get_key()]
                second_type = children[1][self.get_key()]

                # VariableDefinitionStatement is used by solc 0.4.0-0.4.6
                # it's changed in 0.4.7 to VariableDeclarationStatement
                if first_type in ["VariableDefinitionStatement", "VariableDeclarationStatement"]:
                    # only the pre statement can be a variable declaration

                    if len(children) == 2:
                        # only one child apart from body, it must be pre
                        pre, cond, post = children[0], None, None
                    else:
                        # more than one child, figure out which one is the cond
                        if second_type == "ExpressionStatement":
                            # only the post can be an expression statement
                            pre, cond, post = children[0], None, children[1]
                        else:
                            # similarly, the post cannot be anything other than an expression statement
                            pre, cond, post = children[0], children[1], None
                elif first_type == "ExpressionStatement":
                    # the first element can either be pre or post

                    if len(children) == 2:
                        # this is entirely ambiguous, so apply a very dumb heuristic:
                        # if the statement is closer to the start of the body, it's probably the post
                        # otherwise, it's probably the pre
                        # this will work in all cases where the formatting isn't completely borked

                        node_len = int(children[0]["src"].split(":")[1])

                        node_start = int(children[0]["src"].split(":")[0])
                        node_end = node_start + node_len

                        for_start = int(statement["src"].split(":")[0]) + 3  # trim off the 'for'
                        body_start = int(body["src"].split(":")[0])

                        dist_start = node_start - for_start
                        dist_end = body_start - node_end
                        if dist_start > dist_end:
                            pre, cond, post = None, None, children[0]
                        else:
                            pre, cond, post = children[0], None, None
                    else:
                        # more than one child, we must be the pre
                        pre, cond, post = children[0], children[1], None
                else:
                    # the first element must be the cond

                    if len(children) == 2:
                        pre, cond, post = None, children[0], None
                    else:
                        pre, cond, post = None, children[0], children[1]

        return pre, cond, post, body

    def _parse_for(self, statement: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        # ForStatement = 'for' '(' (SimpleStatement)? ';' (Expression)? ';' (ExpressionStatement)? ')' Statement

        if self.is_compact_ast:
            pre, cond, post, body = self._parse_for_compact_ast(statement)
        else:
            pre, cond, post, body = self._parse_for_legacy_ast(statement)

        node_startLoop = self._new_node(NodeType.STARTLOOP, statement["src"], scope)
        node_endLoop = self._new_node(NodeType.ENDLOOP, statement["src"], scope)

        last_scope = scope

        if pre:
            pre_scope = Scope(scope.is_checked, False, last_scope)
            last_scope = pre_scope
            node_init_expression = self._parse_statement(pre, node, pre_scope)
            link_underlying_nodes(node_init_expression, node_startLoop)
        else:
            link_underlying_nodes(node, node_startLoop)

        if cond:
            cond_scope = Scope(scope.is_checked, False, last_scope)
            last_scope = cond_scope
            node_condition = self._new_node(NodeType.IFLOOP, cond["src"], cond_scope)
            node_condition.add_unparsed_expression(cond)
            link_underlying_nodes(node_startLoop, node_condition)

            node_beforeBody = node_condition
        else:
            node_condition = None
            node_beforeBody = node_startLoop

        body_scope = Scope(scope.is_checked, False, last_scope)
        last_scope = body_scope
        node_body = self._parse_statement(body, node_beforeBody, body_scope)

        if post:
            node_loopexpression = self._parse_statement(post, node_body, last_scope)
            link_underlying_nodes(node_loopexpression, node_beforeBody)
        else:
            # node_loopexpression = None
            link_underlying_nodes(node_body, node_beforeBody)

        if node_condition:
            link_underlying_nodes(node_condition, node_endLoop)
        else:
            link_underlying_nodes(
                node_startLoop, node_endLoop
            )  # this is an infinite loop but we can't break our cfg

        return node_endLoop

    def _parse_dowhile(self, do_while_statement: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        node_startDoWhile = self._new_node(NodeType.STARTLOOP, do_while_statement["src"], scope)
        condition_scope = Scope(scope.is_checked, False, scope)

        if self.is_compact_ast:
            node_condition = self._new_node(
                NodeType.IFLOOP, do_while_statement["condition"]["src"], condition_scope
            )
            node_condition.add_unparsed_expression(do_while_statement["condition"])
            statement = self._parse_statement(
                do_while_statement["body"], node_condition, condition_scope
            )
        else:
            children = do_while_statement[self.get_children("children")]
            # same order in the AST as while
            expression = children[0]
            node_condition = self._new_node(NodeType.IFLOOP, expression["src"], condition_scope)
            node_condition.add_unparsed_expression(expression)
            statement = self._parse_statement(children[1], node_condition, condition_scope)

        body_scope = Scope(scope.is_checked, False, condition_scope)
        node_endDoWhile = self._new_node(NodeType.ENDLOOP, do_while_statement["src"], body_scope)

        link_underlying_nodes(node, node_startDoWhile)
        # empty block, loop from the start to the condition
        if not node_condition.underlying_node.sons:
            link_underlying_nodes(node_startDoWhile, node_condition)
        else:
            link_nodes(
                node_startDoWhile.underlying_node,
                node_condition.underlying_node.sons[0],
            )
        link_underlying_nodes(statement, node_condition)
        link_underlying_nodes(node_condition, node_endDoWhile)
        return node_endDoWhile

    def _construct_try_expression(self, externalCall: dict, parameters_list: dict) -> dict:
        # if the parameters are more than 1 we make the leftHandSide of the Assignment node
        # a TupleExpression otherwise an Identifier

        # case when there isn't returns(...)
        # e.g. external call that doesn't have any return variable
        if not parameters_list:
            return externalCall

        ret: dict = {"nodeType": "Assignment", "operator": "=", "src": parameters_list["src"]}

        parameters = parameters_list.get("parameters")

        # if the name is "" it means the return variable is not used
        if len(parameters) == 1:
            if parameters[0]["name"] != "":
                self._add_param(parameters[0])
                ret["typeDescriptions"] = {
                    "typeString": parameters[0]["typeName"]["typeDescriptions"]["typeString"]
                }
                leftHandSide = {
                    "name": parameters[0]["name"],
                    "nodeType": "Identifier",
                    "src": parameters[0]["src"],
                    "referencedDeclaration": parameters[0]["id"],
                    "typeDescriptions": parameters[0]["typeDescriptions"],
                }
            else:
                # we don't need an Assignment so we return only the external call
                return externalCall
        else:
            ret["typeDescriptions"] = {"typeString": "tuple()"}
            leftHandSide = {
                "components": [],
                "nodeType": "TupleExpression",
                "src": parameters_list["src"],
            }

            for i, p in enumerate(parameters):
                if p["name"] == "":
                    continue

                new_statement = {
                    "nodeType": "VariableDefinitionStatement",
                    "src": p["src"],
                    "declarations": [p],
                }
                self._add_param_init_tuple(new_statement, i)

                ident = {
                    "name": p["name"],
                    "nodeType": "Identifier",
                    "src": p["src"],
                    "referencedDeclaration": p["id"],
                    "typeDescriptions": p["typeDescriptions"],
                }
                leftHandSide["components"].append(ident)

        ret["leftHandSide"] = leftHandSide
        ret["rightHandSide"] = externalCall

        return ret

    def _parse_try_catch(self, statement: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        externalCall = statement.get("externalCall")

        if externalCall is None:
            raise ParsingError(f"Try/Catch not correctly parsed by Slither {statement}")
        catch_scope = Scope(scope.is_checked, False, scope)
        new_node = self._new_node(NodeType.TRY, statement["src"], catch_scope)
        clauses = statement.get("clauses", [])
        # the first clause is the try scope
        returned_variables = clauses[0].get("parameters", None)
        constructed_try_expression = self._construct_try_expression(
            externalCall, returned_variables
        )
        new_node.add_unparsed_expression(constructed_try_expression)
        link_underlying_nodes(node, new_node)
        node = new_node

        for index, clause in enumerate(clauses):
            # clauses after the first one are related to catch cases
            # we set the parameters (e.g. data in this case. catch(string memory data) ...)
            # to be initialized so they are not reported by the uninitialized-local-variables detector
            if index >= 1:
                self._parse_catch(clause, node, catch_scope, True)
            else:
                # the parameters for the try scope were already added in _construct_try_expression
                self._parse_catch(clause, node, catch_scope, False)
        return node

    def _parse_catch(
        self, statement: dict, node: NodeSolc, scope: Scope, add_param: bool
    ) -> NodeSolc:
        block = statement.get("block")

        if block is None:
            raise ParsingError(f"Catch not correctly parsed by Slither {statement}")
        try_scope = Scope(scope.is_checked, False, scope)

        try_node = self._new_node(NodeType.CATCH, statement["src"], try_scope)
        link_underlying_nodes(node, try_node)

        if add_param:
            if self.is_compact_ast:
                params = statement.get("parameters")
            else:
                params = statement[self.get_children("children")]

            if params:
                for param in params.get("parameters", []):
                    assert param[self.get_key()] == "VariableDeclaration"
                    self._add_param(param, True)

        return self._parse_statement(block, try_node, try_scope)

    def _parse_variable_definition(self, statement: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        try:
            local_var = LocalVariable()
            local_var.set_function(self._function)
            local_var.set_offset(statement["src"], self._function.compilation_unit)

            local_var_parser = LocalVariableSolc(local_var, statement)
            self._add_local_variable(local_var_parser)
            # local_var.analyze(self)

            new_node = self._new_node(NodeType.VARIABLE, statement["src"], scope)
            new_node.underlying_node.add_variable_declaration(local_var)
            link_underlying_nodes(node, new_node)
            return new_node
        except MultipleVariablesDeclaration:
            # Custom handling of var (a,b) = .. style declaration
            if self.is_compact_ast:
                variables = statement["declarations"]
                count = len(variables)

                if (
                    statement["initialValue"]["nodeType"] == "TupleExpression"
                    and len(statement["initialValue"]["components"]) == count
                ):
                    inits = statement["initialValue"]["components"]
                    i = 0
                    new_node = node
                    for variable in variables:
                        if variable is None:
                            continue
                        init = inits[i]
                        src = variable["src"]
                        i = i + 1

                        new_statement = {
                            "nodeType": "VariableDefinitionStatement",
                            "src": src,
                            "declarations": [variable],
                            "initialValue": init,
                        }
                        new_node = self._parse_variable_definition(new_statement, new_node, scope)

                else:
                    # If we have
                    # var (a, b) = f()
                    # we can split in multiple declarations, without init
                    # Then we craft one expression that does the assignment
                    variables = []
                    i = 0
                    new_node = node
                    for variable in statement["declarations"]:
                        if variable:
                            src = variable["src"]
                            # Create a fake statement to be consistent
                            new_statement = {
                                "nodeType": "VariableDefinitionStatement",
                                "src": src,
                                "declarations": [variable],
                            }
                            variables.append(variable)

                            new_node = self._parse_variable_definition_init_tuple(
                                new_statement, i, new_node, scope
                            )
                        else:
                            variables.append(None)
                        i = i + 1

                    var_identifiers = []
                    # craft of the expression doing the assignement
                    for v in variables:
                        if v is not None:
                            identifier = {
                                "nodeType": "Identifier",
                                "src": v["src"],
                                "name": v["name"],
                                "referencedDeclaration": v["id"],
                                "typeDescriptions": {
                                    "typeString": v["typeDescriptions"]["typeString"]
                                },
                            }
                            var_identifiers.append(identifier)
                        else:
                            var_identifiers.append(None)

                    tuple_expression = {
                        "nodeType": "TupleExpression",
                        "src": statement["src"],
                        "components": var_identifiers,
                    }

                    expression = {
                        "nodeType": "Assignment",
                        "src": statement["src"],
                        "operator": "=",
                        "type": "tuple()",
                        "leftHandSide": tuple_expression,
                        "rightHandSide": statement["initialValue"],
                        "typeDescriptions": {"typeString": "tuple()"},
                    }
                    node = new_node
                    new_node = self._new_node(NodeType.EXPRESSION, statement["src"], scope)
                    new_node.add_unparsed_expression(expression)
                    link_underlying_nodes(node, new_node)

            else:
                count = 0
                children = statement[self.get_children("children")]
                child = children[0]
                while child[self.get_key()] == "VariableDeclaration":
                    count = count + 1
                    child = children[count]

                assert len(children) == (count + 1)
                tuple_vars = children[count]

                variables_declaration = children[0:count]
                i = 0
                new_node = node
                if tuple_vars[self.get_key()] == "TupleExpression":
                    assert len(tuple_vars[self.get_children("children")]) == count
                    for variable in variables_declaration:
                        init = tuple_vars[self.get_children("children")][i]
                        src = variable["src"]
                        i = i + 1
                        # Create a fake statement to be consistent
                        new_statement = {
                            self.get_key(): "VariableDefinitionStatement",
                            "src": src,
                            self.get_children("children"): [variable, init],
                        }

                        new_node = self._parse_variable_definition(new_statement, new_node, scope)
                else:
                    # If we have
                    # var (a, b) = f()
                    # we can split in multiple declarations, without init
                    # Then we craft one expression that does the assignment
                    assert tuple_vars[self.get_key()] in ["FunctionCall", "Conditional"]
                    variables = []
                    for variable in variables_declaration:
                        src = variable["src"]
                        # Create a fake statement to be consistent
                        new_statement = {
                            self.get_key(): "VariableDefinitionStatement",
                            "src": src,
                            self.get_children("children"): [variable],
                        }
                        variables.append(variable)

                        new_node = self._parse_variable_definition_init_tuple(
                            new_statement, i, new_node, scope
                        )
                        i = i + 1
                    var_identifiers = []
                    # craft of the expression doing the assignement
                    for v in variables:
                        identifier = {
                            self.get_key(): "Identifier",
                            "src": v["src"],
                            "attributes": {
                                "value": v["attributes"][self.get_key()],
                                "type": v["attributes"]["type"],
                            },
                        }
                        var_identifiers.append(identifier)

                    expression = {
                        self.get_key(): "Assignment",
                        "src": statement["src"],
                        "attributes": {"operator": "=", "type": "tuple()"},
                        self.get_children("children"): [
                            {
                                self.get_key(): "TupleExpression",
                                "src": statement["src"],
                                self.get_children("children"): var_identifiers,
                            },
                            tuple_vars,
                        ],
                    }
                    node = new_node
                    new_node = self._new_node(NodeType.EXPRESSION, statement["src"], scope)
                    new_node.add_unparsed_expression(expression)
                    link_underlying_nodes(node, new_node)

            return new_node

    def _parse_variable_definition_init_tuple(
        self, statement: dict, index: int, node: NodeSolc, scope
    ) -> NodeSolc:
        local_var = LocalVariableInitFromTuple()
        local_var.set_function(self._function)
        local_var.set_offset(statement["src"], self._function.compilation_unit)

        local_var_parser = LocalVariableInitFromTupleSolc(local_var, statement, index)

        self._add_local_variable(local_var_parser)

        new_node = self._new_node(NodeType.VARIABLE, statement["src"], scope)
        new_node.underlying_node.add_variable_declaration(local_var)
        link_underlying_nodes(node, new_node)
        return new_node

    def _parse_statement(
        self, statement: dict, node: NodeSolc, scope: Scope | Function
    ) -> NodeSolc:
        """

        Return:
            node
        """
        # Statement = IfStatement | WhileStatement | ForStatement | Block | InlineAssemblyStatement |
        #            ( DoWhileStatement | PlaceholderStatement | Continue | Break | Return |
        #                          Throw | EmitStatement | SimpleStatement ) ';'
        # SimpleStatement = VariableDefinition | ExpressionStatement

        name = statement[self.get_key()]
        # SimpleStatement = VariableDefinition | ExpressionStatement
        if name == "IfStatement":
            node = self._parse_if(statement, node, scope)
        elif name == "WhileStatement":
            node = self._parse_while(statement, node, scope)
        elif name == "ForStatement":
            node = self._parse_for(statement, node, scope)
        elif name == "Block":
            node = self._parse_block(statement, node, scope)
        elif name == "UncheckedBlock":
            node = self._parse_unchecked_block(statement, node, scope)
        elif name == "InlineAssembly":
            # Added with solc 0.6 - the yul code is an AST
            if "AST" in statement and not self.compilation_unit.core.skip_assembly:
                self._function.contains_assembly = True
                yul_object = self._new_yul_block(statement["src"], scope)
                entrypoint = yul_object.entrypoint
                exitpoint = yul_object.convert(statement["AST"])

                # technically, entrypoint and exitpoint are YulNodes and we should be returning a NodeSolc here
                # but they both expose an underlying_node so oh well
                link_underlying_nodes(node, entrypoint)
                end_assembly = self._new_node(NodeType.ENDASSEMBLY, statement["src"], scope)
                link_underlying_nodes(exitpoint, end_assembly)
                node = end_assembly
            else:
                asm_node = self._new_node(NodeType.ASSEMBLY, statement["src"], scope)
                self._function.contains_assembly = True
                # Added with solc 0.4.12
                if "operations" in statement:
                    asm_node.underlying_node.add_inline_asm(statement["operations"])
                link_underlying_nodes(node, asm_node)
                end_assembly = self._new_node(NodeType.ENDASSEMBLY, statement["src"], scope)
                link_underlying_nodes(asm_node, end_assembly)
                node = end_assembly
        elif name == "DoWhileStatement":
            node = self._parse_dowhile(statement, node, scope)
        # For Continue / Break / Return / Throw
        # The is fixed later
        elif name == "Continue":
            continue_node = self._new_node(NodeType.CONTINUE, statement["src"], scope)
            link_underlying_nodes(node, continue_node)
            node = continue_node
        elif name == "Break":
            break_node = self._new_node(NodeType.BREAK, statement["src"], scope)
            link_underlying_nodes(node, break_node)
            node = break_node
        elif name == "Return":
            return_node = self._new_node(NodeType.RETURN, statement["src"], scope)
            link_underlying_nodes(node, return_node)
            if self.is_compact_ast:
                if statement.get("expression"):
                    return_node.add_unparsed_expression(statement["expression"])
            else:
                if (
                    self.get_children("children") in statement
                    and statement[self.get_children("children")]
                ):
                    assert len(statement[self.get_children("children")]) == 1
                    expression = statement[self.get_children("children")][0]
                    return_node.add_unparsed_expression(expression)
            node = return_node
        elif name == "Throw":
            throw_node = self._new_node(NodeType.THROW, statement["src"], scope)
            link_underlying_nodes(node, throw_node)
            node = throw_node
        elif name == "EmitStatement":
            # expression = parse_expression(statement[self.get_children('children')][0], self)
            if self.is_compact_ast:
                expression = statement["eventCall"]
            else:
                expression = statement[self.get_children("children")][0]
            new_node = self._new_node(NodeType.EXPRESSION, statement["src"], scope)
            new_node.add_unparsed_expression(expression)
            link_underlying_nodes(node, new_node)
            node = new_node
        elif name in ["VariableDefinitionStatement", "VariableDeclarationStatement"]:
            node = self._parse_variable_definition(statement, node, scope)
        elif name == "ExpressionStatement":
            # assert len(statement[self.get_children('expression')]) == 1
            # assert not 'attributes' in statement
            # expression = parse_expression(statement[self.get_children('children')][0], self)
            if self.is_compact_ast:
                expression = statement[self.get_children("expression")]
            else:
                expression = statement[self.get_children("expression")][0]
            new_node = self._new_node(NodeType.EXPRESSION, statement["src"], scope)
            new_node.add_unparsed_expression(expression)
            link_underlying_nodes(node, new_node)
            node = new_node
        elif name == "TryStatement":
            node = self._parse_try_catch(statement, node, scope)
        # elif name == 'TryCatchClause':
        #     self._parse_catch(statement, node)
        elif name == "RevertStatement":
            if self.is_compact_ast:
                expression = statement[self.get_children("errorCall")]
            else:
                expression = statement[self.get_children("errorCall")][0]
            new_node = self._new_node(NodeType.EXPRESSION, statement["src"], scope)
            new_node.add_unparsed_expression(expression)
            link_underlying_nodes(node, new_node)
            node = new_node
        else:
            raise ParsingError(f"Statement not parsed {name}")

        return node

    def _parse_block(self, block: dict, node: NodeSolc, scope: Scope) -> NodeSolc:
        """
        Return:
            Node
        """
        assert block[self.get_key()] == "Block"

        if self.is_compact_ast:
            statements = block["statements"]
        else:
            statements = block[self.get_children("children")]

        new_scope = Scope(scope.is_checked, False, scope)
        for statement in statements:
            node = self._parse_statement(statement, node, new_scope)
        return node

    def _parse_unchecked_block(self, block: dict, node: NodeSolc, scope):
        """
        Return:
            Node
        """
        assert block[self.get_key()] == "UncheckedBlock"

        if self.is_compact_ast:
            statements = block["statements"]
        else:
            statements = block[self.get_children("children")]

        new_scope = Scope(False, False, scope)

        for statement in statements:
            node = self._parse_statement(statement, node, new_scope)
        return node

    def _update_reachability(self, node: Node) -> None:
        worklist = [node]
        while worklist:
            current = worklist.pop()
            # fix point
            if not current.is_reachable:
                current.set_is_reachable(True)
                worklist.extend(current.sons)

    def _parse_cfg(self, cfg: dict) -> None:
        assert cfg[self.get_key()] == "Block"

        node = self._new_node(NodeType.ENTRYPOINT, cfg["src"], self.underlying_function)
        self._function.entry_point = node.underlying_node

        if self.is_compact_ast:
            statements = cfg["statements"]
        else:
            statements = cfg[self.get_children("children")]

        if not statements:
            self._function.is_empty = True
        else:
            self._function.is_empty = False
            self._parse_block(cfg, node, self.underlying_function)
            self._remove_incorrect_edges()
            self._remove_alone_endif()

    # endregion
    ###################################################################################
    ###################################################################################
    # region Loops
    ###################################################################################
    ###################################################################################

    def _find_end_loop(self, node: Node, visited: list[Node], counter: int) -> Node | None:
        # counter allows to explore nested loop
        if node in visited:
            return None

        if node.type == NodeType.ENDLOOP:
            if counter == 0:
                return node
            counter -= 1

        # nested loop
        if node.type == NodeType.STARTLOOP:
            counter += 1

        visited = visited + [node]
        for son in node.sons:
            ret = self._find_end_loop(son, visited, counter)
            if ret:
                return ret

        return None

    def _find_if_loop(self, node: Node, visited: list[Node], skip_if_loop: int) -> Node | None:
        if node in visited:
            return None

        # If skip_if_loop is not 0 it means we are in a nested situation
        # and we have to skip the closer n loop headers.
        # If in the fathers there is an EXPRESSION node it's a for loop
        # and it's the index increment expression
        if node.type == NodeType.IFLOOP:
            if skip_if_loop == 0:
                for father in node.fathers:
                    if father.type == NodeType.EXPRESSION:
                        return father
                return node

        # skip_if_loop works as explained above.
        # This handle when a for loop doesn't have a condition
        # e.g. for (;;) {}
        # and decrement skip_if_loop since we are out of the nested loop
        if node.type == NodeType.STARTLOOP:
            if skip_if_loop == 0:
                return node
            skip_if_loop -= 1

        # If we find an ENDLOOP means we are in a nested loop
        # we increment skip_if_loop counter
        if node.type == NodeType.ENDLOOP:
            skip_if_loop += 1

        visited = visited + [node]
        for father in node.fathers:
            ret = self._find_if_loop(father, visited, skip_if_loop)
            if ret:
                return ret

        return None

    def _fix_break_node(self, node: Node) -> None:
        end_node = self._find_end_loop(node, [], 0)

        if not end_node:
            # If there is not end condition on the loop
            # The exploration will reach a STARTLOOP before reaching the endloop
            # We start with -1 as counter to catch this corner case
            end_node = self._find_end_loop(node, [], -1)
            if not end_node:
                raise ParsingError(f"Break in no-loop context {node.function}")

        for son in node.sons:
            son.remove_father(node)
        node.set_sons([end_node])
        end_node.add_father(node)

    def _fix_continue_node(self, node: Node) -> None:
        if_loop_node = self._find_if_loop(node, [], 0)

        if not if_loop_node:
            raise ParsingError(f"Continue in no-loop context {node.node_id}")

        for son in node.sons:
            son.remove_father(node)
        node.set_sons([if_loop_node])
        if_loop_node.add_father(node)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Try-Catch
    ###################################################################################
    ###################################################################################

    def _fix_try(self, node: Node) -> None:
        end_node = next((son for son in node.sons if son.type != NodeType.CATCH), None)
        if end_node:
            for son in node.sons:
                if son.type == NodeType.CATCH:
                    self._fix_catch(son, end_node, set())

    def _fix_catch(self, node: Node, end_node: Node, visited: set[Node]) -> None:
        if not node.sons:
            link_nodes(node, end_node)
        else:
            for son in node.sons:
                # If the son is a TRY node it will be fixed later when _fix_try is called with that node
                # otherwise we try to fix it multiple times. It can happen in the case of nested try-catch blocks.
                if son != end_node and son not in visited and son.type != NodeType.TRY:
                    visited.add(son)
                    self._fix_catch(son, end_node, visited)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Params, Returns, Modifiers
    ###################################################################################
    ###################################################################################

    def _add_param(self, param: dict, initialized: bool = False) -> LocalVariableSolc:
        local_var = LocalVariable()
        local_var.set_function(self._function)
        local_var.set_offset(param["src"], self._function.compilation_unit)

        local_var_parser = LocalVariableSolc(local_var, param)

        local_var_parser.analyze(self)

        if initialized:
            local_var.initialized = True

        # see https://solidity.readthedocs.io/en/v0.4.24/types.html?highlight=storage%20location#data-location
        if local_var.location == "default":
            local_var.set_location("memory")

        self._add_local_variable(local_var_parser)
        return local_var_parser

    def _add_param_init_tuple(self, statement: dict, index: int) -> LocalVariableInitFromTupleSolc:
        local_var = LocalVariableInitFromTuple()
        local_var.set_function(self._function)
        local_var.set_offset(statement["src"], self._function.compilation_unit)

        local_var_parser = LocalVariableInitFromTupleSolc(local_var, statement, index)

        self._add_local_variable(local_var_parser)
        return local_var_parser

    def _parse_params(self, params: dict):
        assert params[self.get_key()] == "ParameterList"

        self._function.parameters_src().set_offset(params["src"], self._function.compilation_unit)

        if self.is_compact_ast:
            params = params["parameters"]
        else:
            params = params[self.get_children("children")]

        for param in params:
            assert param[self.get_key()] == "VariableDeclaration"
            local_var = self._add_param(param)
            self._function.add_parameters(local_var.underlying_variable)

    def _parse_returns(self, returns: dict):
        assert returns[self.get_key()] == "ParameterList"

        self._function.returns_src().set_offset(returns["src"], self._function.compilation_unit)

        if self.is_compact_ast:
            self._returnsNotParsed = returns["parameters"]
        else:
            self._returnsNotParsed = returns[self.get_children("children")]

        for ret in self._returnsNotParsed:
            assert ret[self.get_key()] == "VariableDeclaration"
            local_var = self._add_param(ret)
            self._function.add_return(local_var.underlying_variable)

    def _parse_modifier(self, modifier: dict):
        m = parse_expression(modifier, self)
        # self._expression_modifiers.append(m)

        # Do not parse modifier nodes for interfaces
        if not self._function.is_implemented:
            return

        for m in ExportValues(m).result():
            if isinstance(m, Modifier):
                node_parser = self._new_node(
                    NodeType.EXPRESSION, modifier["src"], self.underlying_function
                )
                node_parser.add_unparsed_expression(modifier)
                # The latest entry point is the entry point, or the latest modifier call
                if self._function.modifiers:
                    latest_entry_point = self._function.modifiers_statements[-1].nodes[-1]
                else:
                    latest_entry_point = self._function.entry_point
                insert_node(latest_entry_point, node_parser.underlying_node)
                self._function.add_modifier(
                    ModifierStatements(
                        modifier=m,
                        entry_point=latest_entry_point,
                        nodes=[latest_entry_point, node_parser.underlying_node],
                    )
                )

            elif isinstance(m, Contract):
                node_parser = self._new_node(
                    NodeType.EXPRESSION, modifier["src"], self.underlying_function
                )
                node_parser.add_unparsed_expression(modifier)
                # The latest entry point is the entry point, or the latest constructor call
                if self._function.explicit_base_constructor_calls_statements:
                    latest_entry_point = self._function.explicit_base_constructor_calls_statements[
                        -1
                    ].nodes[-1]
                else:
                    latest_entry_point = self._function.entry_point
                insert_node(latest_entry_point, node_parser.underlying_node)
                self._function.add_explicit_base_constructor_calls_statements(
                    ModifierStatements(
                        modifier=m,
                        entry_point=latest_entry_point,
                        nodes=[latest_entry_point, node_parser.underlying_node],
                    )
                )

    def _fix_implicit_return(self, return_params: dict) -> None:
        """
        Creates an artificial return node iff a function has a named return variable declared in its signature.
        Finds all leaf nodes in the CFG which are not return nodes, and links them to the artificial return node.
        """
        does_not_have_return_params = len(self.underlying_function.returns) == 0
        does_not_have_named_returns = all(
            ret.name == "" for ret in self.underlying_function.returns
        )
        not_implemented = not self._function.is_implemented

        if does_not_have_return_params or does_not_have_named_returns or not_implemented:
            return

        return_node = self._new_node(
            NodeType.RETURN, return_params["src"], self.underlying_function
        )
        for node, node_solc in self._node_to_nodesolc.items():
            if len(node.sons) == 0 and node.type not in [NodeType.RETURN, NodeType.THROW]:
                link_underlying_nodes(node_solc, return_node)

        for _, yul_block in self._node_to_yulobject.items():
            for yul_node in yul_block.nodes:
                node = yul_node.underlying_node
                if len(node.sons) == 0 and node.type not in [NodeType.RETURN, NodeType.THROW]:
                    link_underlying_nodes(yul_node, return_node)

        if self.is_compact_ast:
            self._add_return_exp_compact(return_node, return_params)
        else:
            self._add_return_exp_legacy(return_node, return_params)

        return_node.analyze_expressions(self)

    def _add_return_exp_compact(self, return_node: NodeSolc, return_params: dict) -> None:
        if len(self.underlying_function.returns) == 1:
            return_arg = self.underlying_function.returns[0]
            if return_arg.name != "":
                (refId, refSrc, refType) = next(
                    (ret["id"], ret["src"], ret["typeDescriptions"])
                    for ret in self._returnsNotParsed
                    if ret["name"] == return_arg.name
                )
                return_node.add_unparsed_expression(
                    {
                        "name": return_arg.name,
                        "nodeType": "Identifier",
                        "overloadedDeclarations": [],
                        "referencedDeclaration": refId,
                        "src": refSrc,
                        "typeDescriptions": refType,
                    }
                )
        else:
            expression = {
                "components": [],
                "isConstant": False,
                "isInlineArray": False,
                "isLValue": False,
                "isPure": False,
                "lValueRequested": False,
                "nodeType": "TupleExpression",
                "src": return_params["src"],
                "typeDescriptions": {},
            }
            type_ids = []
            type_strs = []
            for return_arg in self.underlying_function.returns:
                # For each named return variable, we add an identifier to the tuple.
                if return_arg.name != "":
                    (refId, refSrc, refType) = next(
                        (ret["id"], ret["src"], ret["typeDescriptions"])
                        for ret in self._returnsNotParsed
                        if ret["name"] == return_arg.name
                    )
                    type_ids.append(refType["typeIdentifier"])
                    type_strs.append(refType["typeString"])
                    expression["components"].append(
                        {
                            "name": return_arg.name,
                            "nodeType": "Identifier",
                            "overloadedDeclarations": [],
                            "referencedDeclaration": refId,
                            "src": refSrc,
                            "typeDescriptions": refType,
                        }
                    )
            expression["typeDescriptions"]["typeIdentifier"] = (
                "t_tuple$_" + "_$_".join(type_ids) + "_$"
            )
            expression["typeDescriptions"]["typeString"] = "tuple(" + ",".join(type_strs) + ")"
            return_node.add_unparsed_expression(expression)

    def _add_return_exp_legacy(self, return_node: NodeSolc, return_params: dict) -> None:
        if len(self.underlying_function.returns) == 1:
            return_arg = self.underlying_function.returns[0]
            if return_arg.name != "":
                (refSrc, refType) = next(
                    (ret["src"], ret["attributes"]["type"])
                    for ret in self._returnsNotParsed
                    if ret["attributes"]["name"] == return_arg.name
                )
                return_node.add_unparsed_expression(
                    {
                        "attributes": {"type": refType, "value": return_arg.name},
                        "name": "Identifier",
                        "src": refSrc,
                    }
                )
        else:
            expression = {
                "children": [],
                "name": "TupleExpression",
                "src": return_params["src"],
            }
            for return_arg in self.underlying_function.returns:
                # For each named return variable, we add an identifier to the tuple.
                if return_arg.name != "":
                    (refSrc, refType) = next(
                        (ret["src"], ret["attributes"]["type"])
                        for ret in self._returnsNotParsed
                        if ret["attributes"]["name"] == return_arg.name
                    )
                    expression["children"].append(
                        {
                            "attributes": {"type": refType, "value": return_arg.name},
                            "name": "Identifier",
                            "src": refSrc,
                        }
                    )
            return_node.add_unparsed_expression(expression)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Edges
    ###################################################################################
    ###################################################################################

    def _remove_incorrect_edges(self):
        for node in self._node_to_nodesolc:
            if node.type in [NodeType.RETURN, NodeType.THROW]:
                for son in node.sons:
                    son.remove_father(node)
                node.set_sons([])
            if node.type in [NodeType.BREAK]:
                self._fix_break_node(node)
            if node.type in [NodeType.CONTINUE]:
                self._fix_continue_node(node)
            if node.type in [NodeType.TRY]:
                self._fix_try(node)

        # this step needs to happen after all of the break statements are fixed
        # really, we should be passing some sort of context down so the break statement doesn't
        # need to be fixed out-of-band in the first place
        for node in self._node_to_nodesolc:
            if node.type in [NodeType.STARTLOOP]:
                # can we prune? only if after pruning, we have at least one son that isn't itself
                if (
                    len([son for son in node.sons if son.type != NodeType.ENDLOOP and son != node])
                    == 0
                ):
                    continue

                new_sons = []
                for son in node.sons:
                    if son.type != NodeType.ENDLOOP:
                        new_sons.append(son)
                        continue
                    son.remove_father(node)
                node.set_sons(new_sons)

    def _remove_alone_endif(self) -> None:
        """
        Can occur on:
        if(..){
            return
        }
        else{
            return
        }

        Iterate until a fix point to remove the ENDIF node
        creates on the following pattern
        if(){
            return
        }
        else if(){
            return
        }
        """
        prev_nodes = []
        while set(prev_nodes) != set(self._node_to_nodesolc.keys()):
            prev_nodes = self._node_to_nodesolc.keys()
            to_remove: list[Node] = []
            for node in self._node_to_nodesolc:
                if node.type == NodeType.ENDIF and not node.fathers:
                    for son in node.sons:
                        son.remove_father(node)
                    node.set_sons([])
                    to_remove.append(node)
            self._function.nodes = [n for n in self._function.nodes if n not in to_remove]
            for remove in to_remove:
                if remove in self._node_to_nodesolc:
                    del self._node_to_nodesolc[remove]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Ternary
    ###################################################################################
    ###################################################################################

    def _rewrite_ternary_as_if_else(self) -> bool:
        ternary_found = True
        updated = False
        while ternary_found:
            ternary_found = False
            for node in self._node_to_nodesolc:
                has_cond = HasConditional(node.expression)
                if has_cond.result():
                    st = SplitTernaryExpression(node.expression)
                    condition = st.condition
                    if not condition:
                        raise ParsingError(
                            f"Incorrect ternary conversion {node.expression} {node.source_mapping}"
                        )
                    true_expr = st.true_expression
                    false_expr = st.false_expression
                    self._split_ternary_node(node, condition, true_expr, false_expr)
                    ternary_found = True
                    updated = True
                    break
        return updated

    def _split_ternary_node(
        self,
        node: Node,
        condition: "Expression",
        true_expr: "Expression",
        false_expr: "Expression",
    ):
        condition_node = self._new_node(NodeType.IF, node.source_mapping, node.scope)
        condition_node.underlying_node.add_expression(condition)
        condition_node.analyze_expressions(self)

        if node.type == NodeType.VARIABLE:
            condition_node.underlying_node.add_variable_declaration(node.variable_declaration)

        true_node_parser = self._new_node(NodeType.EXPRESSION, node.source_mapping, node.scope)
        if node.type == NodeType.VARIABLE:
            assert isinstance(true_expr, AssignmentOperation)
            # true_expr = true_expr.expression_right
        elif node.type == NodeType.RETURN:
            true_node_parser.underlying_node.type = NodeType.RETURN
        true_node_parser.underlying_node.add_expression(true_expr)
        true_node_parser.analyze_expressions(self)

        false_node_parser = self._new_node(NodeType.EXPRESSION, node.source_mapping, node.scope)
        if node.type == NodeType.VARIABLE:
            assert isinstance(false_expr, AssignmentOperation)
        elif node.type == NodeType.RETURN:
            false_node_parser.underlying_node.type = NodeType.RETURN
            # false_expr = false_expr.expression_right
        false_node_parser.underlying_node.add_expression(false_expr)
        false_node_parser.analyze_expressions(self)

        endif_node = self._new_node(NodeType.ENDIF, node.source_mapping, node.scope)

        for father in node.fathers:
            father.replace_son(node, condition_node.underlying_node)
            condition_node.underlying_node.add_father(father)

        for son in node.sons:
            son.remove_father(node)
            son.add_father(endif_node.underlying_node)
            endif_node.underlying_node.add_son(son)

        link_underlying_nodes(condition_node, true_node_parser)
        link_underlying_nodes(condition_node, false_node_parser)

        if true_node_parser.underlying_node.type not in [
            NodeType.THROW,
            NodeType.RETURN,
        ]:
            link_underlying_nodes(true_node_parser, endif_node)
        if false_node_parser.underlying_node.type not in [
            NodeType.THROW,
            NodeType.RETURN,
        ]:
            link_underlying_nodes(false_node_parser, endif_node)

        self._function.nodes = [n for n in self._function.nodes if n.node_id != node.node_id]
        del self._node_to_nodesolc[node]

    # endregion



from slither import Slither
from slither.core.declarations import Function, FunctionContract
from slither.core.slither_core import SlitherCore


class ResolveFunctionException(Exception):
    pass


def resolve_function(slither: SlitherCore, contract_name: str, function_name: str) -> Function:
    """
    Resolves a function instance, given a contract name and function.
    :param contract_name: The name of the contract the function is declared in.
    :param function_name: The name of the function to resolve.
    :return: Returns the resolved function, raises an exception otherwise.
    """
    # Obtain the target contract
    contracts = slither.get_contract_from_name(contract_name)

    # Verify the contract was resolved successfully
    if len(contracts) != 1:
        raise ResolveFunctionException(f"Could not resolve target contract: {contract_name}")
    contract = contracts[0]
    # Obtain the target function
    target_function = next(
        (function for function in contract.functions if function.name == function_name),
        None,
    )

    # Verify we have resolved the function specified.
    if target_function is None:
        raise ResolveFunctionException(
            f"Could not resolve target function: {contract_name}.{function_name}"
        )

    # Add the resolved function to the new list.
    return target_function


def resolve_functions(slither: Slither, functions: list[str | tuple[str, str]]) -> list[Function]:
    """
    Resolves the provided function descriptors.
    :param functions: A list of tuples (contract_name, function_name) or str (of form "ContractName.FunctionName")
    to resolve into function objects.
    :return: Returns a list of resolved functions.
    """
    # Create the resolved list.
    resolved: list[Function] = []

    # Verify that the provided argument is a list.
    if not isinstance(functions, list):
        raise ResolveFunctionException("Provided functions to resolve must be a list type.")

    # Loop for each item in the list.
    for item in functions:
        if isinstance(item, str):
            # If the item is a single string, we assume it is of form 'ContractName.FunctionName'.
            parts = item.split(".")
            if len(parts) < 2:
                raise ResolveFunctionException(
                    "Provided string descriptor must be of form 'ContractName.FunctionName'"
                )
            resolved.append(resolve_function(slither, parts[0], parts[1]))
        elif isinstance(item, tuple):
            # If the item is a tuple, it should be a 2-tuple providing contract and function names.
            if len(item) != 2:
                raise ResolveFunctionException(
                    "Provided tuple descriptor must provide a contract and function name."
                )
            resolved.append(resolve_function(slither, item[0], item[1]))
        else:
            raise ResolveFunctionException(
                f"Unexpected function descriptor type to resolve in list: {type(item)}"
            )

    # Return the resolved list.
    return resolved


def all_function_definitions(function: Function) -> list[Function]:
    """
    Obtains a list of representing this function and any base definitions
    :param function: The function to obtain all definitions at and beneath.
    :return: Returns a list composed of the provided function definition and any base definitions.
    """
    # TODO implement me
    if not isinstance(function, FunctionContract):
        return []
    ret: list[Function] = [function]
    ret += [
        f
        for c in function.contract.inheritance
        for f in c.functions_and_modifiers_declared
        if f.full_name == function.full_name
    ]
    return ret


def __find_target_paths(
    slither: SlitherCore, target_function: Function, current_path: list[Function] | None = None
) -> set[tuple[Function, ...]]:
    current_path = current_path if current_path else []
    # Create our results list
    results: set[tuple[Function, ...]] = set()

    # Add our current function to the path.
    current_path = [target_function] + current_path

    # Obtain this target function and any base definitions.
    all_target_functions = set(all_function_definitions(target_function))

    # Look through all functions
    for contract in slither.contracts:
        for function in contract.functions_and_modifiers_declared:
            # If the function is already in our path, skip it.
            if function in current_path:
                continue

            # Find all function calls in this function (except for low level)
            called_functions_list = [
                ir.function
                for _, ir in function.high_level_calls
                if isinstance(ir.function, Function)
            ]
            called_functions_list += [ir.function for ir in function.library_calls]
            called_functions_list += [
                ir.function for ir in function.internal_calls if isinstance(ir.function, Function)
            ]
            called_functions = set(called_functions_list)

            # If any of our target functions are reachable from this function, it's a result.
            if all_target_functions.intersection(called_functions):
                path_results = __find_target_paths(slither, function, current_path.copy())
                if path_results:
                    results = results.union(path_results)

    # If this path is external accessible from this point, we add the current path to the list.
    if target_function.visibility in ["public", "external"] and len(current_path) > 1:
        results.add(tuple(current_path))

    return results


def find_target_paths(
    slither: SlitherCore, target_functions: list[Function]
) -> set[tuple[Function, ...]]:
    """
    Obtains all functions which can lead to any of the target functions being called.
    :param target_functions: The functions we are interested in reaching.
    :return: Returns a list of all functions which can reach any of the target_functions.
    """
    # Create our results list
    results: set[tuple[Function, ...]] = set()

    # Loop for each target function
    for target_function in target_functions:
        results = results.union(__find_target_paths(slither, target_function))

    return results



from pathlib import Path
from slither import Slither
from slither.visitors.expression.constants_folding import ConstantFolding

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
CONSTANT_FOLDING_TEST_ROOT = Path(TEST_DATA_DIR, "constant_folding")


def test_constant_folding_unary(solc_binary_path):
    solc_path = solc_binary_path("0.8.0")
    file = Path(CONSTANT_FOLDING_TEST_ROOT, "constant_folding_unary.sol").as_posix()
    Slither(file, solc=solc_path)


def test_constant_folding_rational(solc_binary_path):
    solc_path = solc_binary_path("0.8.0")
    s = Slither(
        Path(CONSTANT_FOLDING_TEST_ROOT, "constant_folding_rational.sol").as_posix(), solc=solc_path
    )
    contract = s.get_contract_from_name("C")[0]

    variable_a = contract.get_state_variable_from_name("a")
    assert str(variable_a.type) == "uint256"
    assert ConstantFolding(variable_a.expression, "uint256").result().value == 10

    variable_b = contract.get_state_variable_from_name("b")
    assert str(variable_b.type) == "int128"
    assert ConstantFolding(variable_b.expression, "int128").result().value == 2

    variable_c = contract.get_state_variable_from_name("c")
    assert str(variable_c.type) == "int64"
    assert ConstantFolding(variable_c.expression, "int64").result().value == 3

    variable_d = contract.get_state_variable_from_name("d")
    assert str(variable_d.type) == "int256"
    assert ConstantFolding(variable_d.expression, "int256").result().value == 1500

    variable_e = contract.get_state_variable_from_name("e")
    assert str(variable_e.type) == "uint256"
    assert (
        ConstantFolding(variable_e.expression, "uint256").result().value
        == 57896044618658097711785492504343953926634992332820282019728792003956564819968
    )

    variable_f = contract.get_state_variable_from_name("f")
    assert str(variable_f.type) == "uint256"
    assert (
        ConstantFolding(variable_f.expression, "uint256").result().value
        == 115792089237316195423570985008687907853269984665640564039457584007913129639935
    )

    variable_g = contract.get_state_variable_from_name("g")
    assert str(variable_g.type) == "int64"
    assert ConstantFolding(variable_g.expression, "int64").result().value == -7


def test_constant_folding_binary_expressions(solc_binary_path):
    sl = Slither(
        Path(CONSTANT_FOLDING_TEST_ROOT, "constant_folding_binop.sol").as_posix(),
        solc=solc_binary_path("0.8.0"),
    )
    contract = sl.get_contract_from_name("BinOp")[0]

    variable_a = contract.get_state_variable_from_name("a")
    assert str(variable_a.type) == "uint256"
    assert ConstantFolding(variable_a.expression, "uint256").result().value == 0

    variable_b = contract.get_state_variable_from_name("b")
    assert str(variable_b.type) == "uint256"
    assert ConstantFolding(variable_b.expression, "uint256").result().value == 3

    variable_c = contract.get_state_variable_from_name("c")
    assert str(variable_c.type) == "uint256"
    assert ConstantFolding(variable_c.expression, "uint256").result().value == 3

    variable_d = contract.get_state_variable_from_name("d")
    assert str(variable_d.type) == "bool"
    assert ConstantFolding(variable_d.expression, "bool").result().value is False

    variable_e = contract.get_state_variable_from_name("e")
    assert str(variable_e.type) == "bool"
    assert ConstantFolding(variable_e.expression, "bool").result().value is False

    variable_f = contract.get_state_variable_from_name("f")
    assert str(variable_f.type) == "bool"
    assert ConstantFolding(variable_f.expression, "bool").result().value is True

    variable_g = contract.get_state_variable_from_name("g")
    assert str(variable_g.type) == "bool"
    assert ConstantFolding(variable_g.expression, "bool").result().value is False

    variable_h = contract.get_state_variable_from_name("h")
    assert str(variable_h.type) == "bool"
    assert ConstantFolding(variable_h.expression, "bool").result().value is False

    variable_i = contract.get_state_variable_from_name("i")
    assert str(variable_i.type) == "bool"
    assert ConstantFolding(variable_i.expression, "bool").result().value is True

    variable_j = contract.get_state_variable_from_name("j")
    assert str(variable_j.type) == "bool"
    assert ConstantFolding(variable_j.expression, "bool").result().value is False

    variable_k = contract.get_state_variable_from_name("k")
    assert str(variable_k.type) == "bool"
    assert ConstantFolding(variable_k.expression, "bool").result().value is True

    variable_l = contract.get_state_variable_from_name("l")
    assert str(variable_l.type) == "uint256"
    assert (
        ConstantFolding(variable_l.expression, "uint256").result().value
        == 115792089237316195423570985008687907853269984665640564039457584007913129639935
    )

    IMPLEMENTATION_SLOT = contract.get_state_variable_from_name("IMPLEMENTATION_SLOT")
    assert str(IMPLEMENTATION_SLOT.type) == "bytes32"
    assert (
        int.from_bytes(
            ConstantFolding(IMPLEMENTATION_SLOT.expression, "bytes32").result().value,
            byteorder="big",
        )
        == 24440054405305269366569402256811496959409073762505157381672968839269610695612
    )

    variable_m = contract.get_state_variable_from_name("m")
    assert str(variable_m.type) == "bytes2"
    assert ConstantFolding(variable_m.expression, "bytes2").result().value == "ab"



from slither.core.declarations import SolidityVariableComposed
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.variable import Variable
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue, is_valid_rvalue, RVALUE, LVALUE
from slither.slithir.variables.reference import ReferenceVariable


class Index(OperationWithLValue):
    def __init__(
        self, result: ReferenceVariable, left_variable: Variable, right_variable: RVALUE
    ) -> None:
        super().__init__()
        assert (
            is_valid_lvalue(left_variable)
            or left_variable == SolidityVariableComposed("msg.data")
            or isinstance(left_variable, TopLevelVariable)
        )
        assert is_valid_rvalue(right_variable)
        assert isinstance(result, ReferenceVariable)
        self._variables = [left_variable, right_variable]
        self._lvalue: ReferenceVariable = result

    @property
    def read(self) -> list[SourceMapping]:
        return list(self.variables)

    @property
    def variables(self) -> list[LVALUE | RVALUE | SolidityVariableComposed]:
        return self._variables  # type: ignore

    @property
    def variable_left(self) -> LVALUE | SolidityVariableComposed:
        return self._variables[0]  # type: ignore

    @property
    def variable_right(self) -> RVALUE:
        return self._variables[1]  # type: ignore

    def __str__(self) -> str:
        return f"{self.lvalue}({self.lvalue.type}) -> {self.variable_left}[{self.variable_right}]"
